from __future__ import annotations

import contextlib
import itertools
from typing import Any, Callable

import deepmerge
import ibis
import ibis.common.exceptions as com
import ibis.expr.operations as ops
import ibis.expr.types as ir
import networkx as nx
import rustworkx as rx
from bidict import bidict
from rich.markup import escape
from rich.text import Text
from rustworkx.visit import DFSVisitor, PruneSearch
from textual.app import ComposeResult
from textual.widgets import Footer

from vinyl.lib.utils.collection import list_union
from vinyl.lib.utils.graphics import TurntableTextualApp, _adjust_type


## Note: outside of DAG class elsewhere, use `to_networkx()` for DAG to avoid errors
def _rustworkx_to_networkx(
    graph: rx.PyDiGraph,
    pre_convert: Callable[..., Any] | None = None,
    post_convert: Callable[..., Any] | None = None,
) -> nx.DiGraph:
    """
    Converts a rustworkx graph to a networkx graph.
    """

    g = nx.DiGraph()
    cur_nodes = graph.nodes()
    if pre_convert is not None:
        cur_nodes = [pre_convert(i) for i in cur_nodes]
    cur_edges = [(cur_nodes[u], cur_nodes[v]) for u, v in graph.edge_list()]
    cur_edge_values = [graph.get_edge_data(u, v) for u, v in graph.edge_list()]
    g.add_nodes_from(nodes_for_adding=cur_nodes)
    for i, edge in enumerate(cur_edges):
        g.add_edge(edge[0], edge[1], cur_edge_values[i])

    if post_convert is not None:
        nx.relabel_nodes(g, post_convert, copy=False)
    return g


## NOTE: Made an attempt at a fix here, but need to verify. Necessary because ops.Join is no longer a valid class
def _get_type(node: ir.Expr) -> Text:
    with contextlib.suppress(AttributeError, NotImplementedError):
        return _adjust_type(node.dtype)

    try:
        schema = node.schema
    except (AttributeError, NotImplementedError):
        # TODO(kszucs): this branch should be removed
        try:
            # As a last resort try get the name of the output_type class
            return node.output_type.__name__
        except (AttributeError, NotImplementedError):
            return Text("\u2205")  # empty set character
    except com.IbisError:
        assert isinstance(node, ops.JoinChain)
        tables = [node.first.parent]
        tables.extend([n.table.parent for n in node.rest])
        table_names = [getattr(t, "name", None) for t in tables]
        schemas = [t.schema for t in tables]
        pairs: list[tuple[str, str]] = []
        for i, table in enumerate(tables):
            pairs.extend(
                (f"{table_names[i]}.{column}", type)
                for column, type in schemas[i].items()
            )
        schema = ibis.schema(pairs)

    out = Text("\n   ")
    len_loop = len(schema.names)
    for i, (name, type) in enumerate(zip(schema.names, schema.types)):
        out += Text(escape(name), style="italic") + Text(": ") + _adjust_type(type)
        if i < len_loop - 1:
            out += Text("\n   ")

    return out


def _get_label(node: ir.Expr, markup=True) -> Text | str:
    typename = _get_type(node)  # Already an escaped string
    name = type(node).__name__
    nodename = (
        node.name
        if isinstance(
            node,
            (
                ops.Literal,
                ops.Field,
                ops.Alias,
                ops.PhysicalTable,
                ops.RangeWindowFrame,
            ),
        )
        else None
    )
    if nodename is not None:
        # [TODO] Don't show nodename because it's too long and ruins the image
        if isinstance(node, ops.window.RangeWindowFrame):
            label = Text(escape(name), style="bold")
        else:
            label = (
                Text(escape(nodename), style="italic")
                + Text(": ")
                + Text(escape(name), style="bold")
            )

            if isinstance(node, ops.Relation):
                label += typename
            else:
                label += Text("\n   :: ") + typename

    else:
        label = Text(escape(name), style="bold")
        if isinstance(node, ops.Relation):
            label += typename
        else:
            label += Text("\n   :: ") + typename
    if markup:
        return label.markup
    return label


class SuccessorVisitor(DFSVisitor):
    def __init__(self, source, depth_limit: int | None = None):
        self.edges: Any = []
        self.degrees = {node: 0 for node in source}
        self.depth_limit = depth_limit

    def tree_edge(self, edge):
        self.edges.append(edge)
        new_degree = self.degrees.get(edge[0]) + 1
        self.degrees[edge[1]] = new_degree
        if new_degree == self.depth_limit:
            raise PruneSearch()


class DAG:
    g: rx.PyDiGraph
    node_dict: bidict[Any, int]

    def __init__(self):
        self.g = rx.PyDiGraph()
        self.node_dict = bidict({})

    def add_node(self, node: Any):
        if node not in self.node_dict:
            p = self.g.add_node(node)
            self.node_dict[node] = p
        return self.node_dict[node]

    def add_edge(self, parent: Any, child: Any, edge: dict[Any, Any] = {}):
        if parent not in self.node_dict:
            self.add_node(parent)
        if child not in self.node_dict:
            self.add_node(child)

        pi = self.node_dict[parent]
        ci = self.node_dict[child]

        if (pi, ci) not in self.g.edge_list():
            parent_node = self.node_dict[parent]
            child_node = self.node_dict[child]
            if parent_node != child_node:
                self.g.add_edge(self.node_dict[parent], self.node_dict[child], edge)
        elif edge is not None:
            old_edge_contents = self.g.get_edge_data(pi, ci)
            if old_edge_contents is None:
                new_edge = deepmerge.always_merger.merge(old_edge_contents, edge)
            else:
                new_edge = edge
            self.g.update_edge(pi, ci, new_edge)

    def get_parents(
        self,
        sources: Any,
    ):
        indexes = [self.node_dict[n] for n in sources]
        out = []
        for i in indexes:
            out.extend(self.g.predecessors(i))

        return list(set(out))

    def get_relatives(
        self,
        sources: Any,
        depth: int | None = None,
        reverse: bool = False,
        include_sources: bool = True,
    ):
        if depth == 0:
            return sources if include_sources else []
        indexes = [self.node_dict[n] for n in sources]
        copy = self.g.copy()
        if reverse:
            copy.reverse()
        g_func = copy if reverse else self.g

        vis = SuccessorVisitor(indexes, depth)
        rx.dfs_search(g_func, indexes, vis)  # type: ignore [attr-defined]
        final_indexes = [k for k, v in vis.degrees.items() if v != 0 or include_sources]
        return [self.node_dict.inv[k] for k in final_indexes]

    def get_ancestors_and_descendants(
        self,
        sources: Any,
        predecessor_depth: int | None,
        successor_depth: int | None,
        include_sources: bool = True,
    ):
        ancestors = self.get_relatives(
            sources, predecessor_depth, reverse=True, include_sources=include_sources
        )
        descendants = self.get_relatives(
            sources, successor_depth, include_sources=include_sources
        )
        return list(set(ancestors + descendants))

    def remove_nodes_and_reconnect(
        self, nodes: list[Any], merge_edge_data: bool = True
    ):
        indexes = [
            n
            for n in self.topological_sort(return_indices=True)
            if self.node_dict.inv[n] in nodes
        ]
        for i in indexes:
            sources = [k for k in self.g.predecessor_indices(i)]
            targets = [k for k in self.g.successor_indices(i)]
            new_edges = list(itertools.product(sources, targets))
            new_edges = [
                (source, target) for source, target in new_edges if source != target
            ]  # remove self-loops

            for source, target in new_edges:
                source_val = self.node_dict.inv[source]
                target_val = self.node_dict.inv[target]
                if merge_edge_data:
                    # Preserve ntypes from A -> B and B -> C when removing node B
                    data_A_B = self.g.get_edge_data(source, i)
                    data_B_C = self.g.get_edge_data(i, target)
                    data_A_C = deepmerge.always_merger.merge(data_A_B, data_B_C)
                    self.add_edge(
                        source_val, target_val, data_A_C
                    )  # merges if edge currently exists
                else:
                    self.add_edge(source_val, target_val)
            self.g.remove_node(i)
            del self.node_dict.inv[i]
        return self.g

    def topological_sort(self, return_indices: bool = False):
        indices = rx.topological_sort(self.g)  # type: ignore [attr-defined]
        if return_indices:
            return indices
        return [self.node_dict.inv[i] for i in indices]

    def subgraph(self, nodes: list[Any]):
        out = DAG()
        out.g = self.g.subgraph([self.node_dict[n] for n in nodes])
        out.node_dict = bidict({v: i for i, v in enumerate(out.g.nodes())})

        return out

    def copy(self) -> DAG:
        new = DAG()
        new.g = self.g.copy()
        new.node_dict = self.node_dict.copy()
        return new

    def to_networkx(self) -> nx.DiGraph:
        g = nx.DiGraph()
        cur_nodes = self.g.nodes()
        edge_list = self.g.edge_list()
        cur_edges = [
            (self.node_dict.inv[u], self.node_dict.inv[v]) for u, v in edge_list
        ]
        cur_edge_values = [self.g.get_edge_data(u, v) for u, v in edge_list]
        g.add_nodes_from(nodes_for_adding=cur_nodes)
        for i, edge in enumerate(cur_edges):
            g.add_edge(edge[0], edge[1], **cur_edge_values[i])

        return g

    def get_node_link_json(self) -> str:
        return nx.node_link_data(self.to_networkx())

    def relabel(self, label_func: Callable[..., Any]) -> DAG:
        cur_edges = self.get_edge_tuples()
        cur_edge_values = [self.g.get_edge_data(u, v) for u, v in self.g.edge_list()]
        new = DAG()
        cur_nodes = self.g.nodes()
        for node in cur_nodes:
            new.add_node(label_func(node))
        for i, edge in enumerate(cur_edges):
            new.add_edge(label_func(edge[0]), label_func(edge[1]), cur_edge_values[i])
        self = new
        return self

    def reverse(self):
        self.g.reverse()
        return self

    def get_edge_tuples(self):
        return [
            (self.node_dict.inv[u], self.node_dict.inv[v])
            for u, v in self.g.edge_list()
        ]

    def get_roots(self):
        return [
            self.node_dict.inv[i]
            for i in self.g.node_indexes()
            if self.g.in_degree(i) == 0
        ]

    def get_ancestors(self, nodes: list[Any]):
        indexes = [self.node_dict[n] for n in nodes]
        out = []
        for i in indexes:
            ancestor_indexes = rx.ancestors(self.g, i)
            out.extend([self.node_dict.inv[k] for k in ancestor_indexes])
        return list(set(out))

    def get_ancestor_roots(self, nodes: list[Any]):
        return list(set(self.get_ancestors(nodes)) & set(self.get_roots()))

    def visualize(self):
        from netext import ConsoleGraph
        from netext.edge_rendering.arrow_tips import ArrowTip
        from netext.edge_rendering.modes import EdgeSegmentDrawingMode
        from netext.edge_routing.modes import EdgeRoutingMode
        from netext.textual.widget import GraphView

        g = self.to_networkx()

        nx.set_node_attributes(
            g,
            {n: n for i, n in enumerate(g.nodes())},
            "$label",
        )

        nx.set_node_attributes(g, lambda n, d, s: d["$label"], "$content-renderer")
        nx.set_edge_attributes(g, EdgeRoutingMode.ORTHOGONAL, "$edge-routing-mode")
        nx.set_edge_attributes(
            g, EdgeSegmentDrawingMode.BOX_ROUNDED, "$edge-segment-drawing-mode"
        )
        nx.set_edge_attributes(g, ArrowTip.ARROW, "$start-arrow-tip")

        if False:
            # console = Console(
            #     width=1000
            # )  # make width large to avoid jupyter cutting off output
            # console.print(ConsoleGraph(g, zoom=1.2, console=console))
            print(ConsoleGraph(g, zoom=1.2))

        else:

            class GraphviewApp(TurntableTextualApp):
                def compose(self) -> ComposeResult:
                    yield GraphView(g)
                    yield Footer()

            GraphviewApp().run()


def nx_remove_node_and_reconnect(
    g: nx.DiGraph, node: str, preserve_ntype: bool = False
) -> nx.DiGraph:
    if g.is_directed():
        sources = [source for source, _ in g.in_edges(node)]
        targets = [target for _, target in g.out_edges(node)]
    else:
        sources = g.neighbors(node)
        targets = g.neighbors(node)

    new_edges = list(itertools.product(sources, targets))
    new_edges = [
        (source, target) for source, target in new_edges if source != target
    ]  # remove self-loops

    if preserve_ntype:
        for source, target in new_edges:
            # Preserve ntypes from A -> B and B -> C when removing node B
            ntype_A_B = g.edges[source, node]["ntype"]
            ntype_B_C = g.edges[node, target]["ntype"]
            ntype_A_C = list_union(ntype_A_B, ntype_B_C)
            if g.has_edge(source, target):
                ntype_A_C = list_union(g.edges[source, target]["ntype"], ntype_A_C)
            else:
                g.add_edge(source, target)
            g.edges[source, target]["ntype"] = ntype_A_C
    else:
        g.add_edges_from(new_edges)
    g.remove_node(node)
    return g



import os
from typing import Any, Callable

import ibis
from ibis.common.exceptions import IbisError

from vinyl.lib.connect import DatabaseFileConnector, _DatabaseConnector, _TableConnector
from vinyl.lib.constants import PreviewHelper
from vinyl.lib.field import Field
from vinyl.lib.settings import _load_project_module
from vinyl.lib.table import VinylTable


def _get_twin_relative_path(name: str, *subdirs) -> str:
    return os.path.join(
        "data",
        *subdirs,
        f"{name}.duckdb",
    )


def source(resource: Callable[..., Any], sample_row_count: int | None = 1000):
    def decorator(cls) -> VinylTable:
        connector = resource()

        schema = []
        for name, attr in cls.__annotations__.items():
            schema.append((name, attr))

        ibis_table = ibis.table(schema, name=cls._unique_name)
        table = VinylTable._create_from_schema(
            ibis.Schema.from_tuples(schema), cls._unique_name
        )
        parent_name = ".".join(cls.__module__.split(".")[2:])

        table._is_vinyl_source = True  # helps find sources in the project for Defs
        table._source_class = cls

        unbounded_op = table.tbl.op()
        if cls._col_replace != {}:
            table._col_replace = {unbounded_op: cls._col_replace}
        else:
            table._col_replace = {}

        if isinstance(connector, _DatabaseConnector):
            # declare here instead of in `full_connection` to avoid cls recursion
            db_conn = str(cls._database)
            schema_conn = str(cls._schema)
            table_conn = str(cls._table)

            def full_connection():
                return connector._get_table(
                    database=db_conn, schema=schema_conn, table=table_conn
                ).op()

            if PreviewHelper._preview == "full":
                table._conn_replace = {unbounded_op: full_connection}
            elif PreviewHelper._preview == "twin":
                # get twin
                module_path = _load_project_module().__file__
                if module_path is None:
                    raise ValueError(
                        "Unable to find the module path for the current project"
                    )
                twin_conn = DatabaseFileConnector(
                    os.path.join(
                        os.path.dirname(module_path),
                        _get_twin_relative_path(resource.__name__),
                    )
                )

                try:
                    table._twin_conn_replace = {
                        unbounded_op: lambda: twin_conn._get_table(
                            database=twin_conn._database,
                            schema=cls._schema,
                            table=cls._table,
                        ).op()
                    }
                except IbisError:
                    # twin doesn't exist, fall back to full conn, but not connected yet
                    table._conn_replace = {unbounded_op: full_connection}

            else:
                raise ValueError(
                    f"Invalid value for PreviewHelper.preview: {PreviewHelper._preview}"
                )
        elif isinstance(connector, _TableConnector):
            # declare here instead of in `full_connection` to avoid cls recursion
            path_conn = str(cls._path)

            def full_connection():
                # for file connectors, we reconnect to the individual file to get the correct table. Since these tables are not in memory, we need to read to get the location.
                adj_conn = type(connector)(path_conn)
                adj_conn._connect()
                tbl_for_op = next(iter(adj_conn._tbls.values()))
                return tbl_for_op.op()

            if PreviewHelper._preview == "full":
                table._conn_replace = {unbounded_op: full_connection}

            elif PreviewHelper._preview == "twin":
                try:
                    table._twin_conn_replace = {
                        unbounded_op: table.tbl.filter(
                            ibis.random()
                            < ibis.least(sample_row_count / cls._row_count, 1)
                        ).op()
                    }
                except IbisError:
                    # twin doesn't exist, fall back to full conn, but not connected yet
                    table._conn_replace = {unbounded_op: full_connection}
        else:
            raise NotImplementedError(
                f"Connector type {type(connector)} is not yet supported"
            )

        for name, attr in cls.__annotations__.items():
            field = Field(
                name=name,
                type=attr,
                parent_ibis_table=ibis_table,
                parent_name=parent_name,
            )
            if hasattr(cls, name):
                current: Field = getattr(cls, name)
                # Field info overrides the class level attributes, may want to change
                field._update(**current.asdict())
            setattr(cls, name, field)

        # update parent table to include the full set of annotations and save graph
        for name, attr in cls.__annotations__.items():
            current = getattr(cls, name)
            # Field info overrides the class level attributes, may want to change
            current._update(**{"parent_vinyl_table": table})

            # Make sure primary key col is set to unique even if user doesn't specify
            if current.primary_key:
                current.unique = True
            setattr(cls, name, current)
            current._update_source_class()
            # make sure final tbl (with source connections) is available for relations graph before storing
            current._store_source_adj_tbl(table)
            current._store_relations()

        return table

    return decorator



import os

import ibis

from vinyl.lib.table import VinylTable


def _return_vinyl_table(cls):
    def wrapper(*args, **kwargs) -> VinylTable:
        adj_path_base = os.path.dirname(os.path.dirname(__file__))
        return VinylTable(
            ibis.read_parquet(
                os.path.join(adj_path_base, cls.path), table_name=cls.__name__
            )._arg
        )

    wrapper.description = cls.description
    wrapper.references = cls.references

    return wrapper


class ExampleDataset:
    description: str
    references: list[str] | None
    path: str


@_return_vinyl_table
class airports(ExampleDataset):
    description = """This dataset lists US airports, including airport code, city, state, latitude, and longitude. This dataset is a subset of the data compiled and published at http://ourairports.com/data/, and is in the public domain."""
    references = None
    path = "vinyl/data/airports.parquet"


@_return_vinyl_table
class anscombe(ExampleDataset):
    description = """Anscombe's Quartet is a famous dataset constructed by Francis Anscombe [1]_. Common summary statistics are identical for each subset of the data, despite the subsets having vastly different characteristics."""
    references = [
        "Anscombe, F. J. (1973). 'Graphs in Statistical Analysis'. American Statistician. 27 (1): 17-21. JSTOR 2682899."
    ]
    path = "vinyl/data/anscombe.parquet"


@_return_vinyl_table
class barley(ExampleDataset):
    description = """This dataset contains crop yields over different regions and different years in the 1930s. It was originally published by Immer in 1934 [1]_, and was popularized by Fisher (1947) [2]_ and Cleveland (1993) [3]_."""
    references = [
        "Immer, F.R., Hayes, H.D. and LeRoy Powers (1934) Statistical determination of barley varietal adaptation. Journal of the American Society for Agronomy 26, 403-419.",
        "Fisher, R.A. (1947) The Design of Experiments. 4th edition. Edinburgh: Oliver and Boyd.",
        "Cleveland, WS (1993) Visualizing data. Hobart Press",
    ]
    path = "vinyl/data/barley.parquet"


@_return_vinyl_table
class birdstrikes(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/birdstrikes.parquet"


@_return_vinyl_table
class budget(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/budget.parquet"


@_return_vinyl_table
class budgets(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/budgets.parquet"


@_return_vinyl_table
class burtin(ExampleDataset):
    description = """This is a famous dataset gathered by Will Burtin to explore the effectiveness of various antibiotics in treating a variety of bacterial infections. The data was first published in 1951; this dataset comes from https://github.com/mbostock/protovis, and is published there under a BSD license."""
    references = None
    path = "vinyl/data/burtin.parquet"


@_return_vinyl_table
class cars(ExampleDataset):
    description = """Acceleration, horsepower, fuel efficiency, weight, and other characteristics of different makes and models of cars. This dataset was originally published by Donoho et al (1982) [1]_, and was made public at http://lib.stat.cmu.edu/datasets/"""
    references = [
        "Donoho, David and Ramos, Ernesto (1982), ``PRIMDATA:  Data Sets for Use With PRIM-H'' (DRAFT)."
    ]
    path = "vinyl/data/cars.parquet"


@_return_vinyl_table
class climate(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/climate.parquet"


@_return_vinyl_table
class co2_concentration(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/co2_concentration.parquet"


@_return_vinyl_table
class countries(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/countries.parquet"


@_return_vinyl_table
class crimea(ExampleDataset):
    description = """This is a dataset containing monthly casualty counts from the Crimean war. It was originally published in 1858 by Florence Nightingale [1]_ in connection with her famous 'Coxcomb' charts."""
    references = [
        "Nightingale, F. (1858) 'Notes on Matters Affecting the Health, Efficiency and Hospital Administration of the British Army'. RCIN 1075240"
    ]
    path = "vinyl/data/crimea.parquet"


@_return_vinyl_table
class disasters(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/disasters.parquet"


@_return_vinyl_table
class driving(ExampleDataset):
    description = """This dataset tracks miles driven per capita along with gas prices annually from 1956 to 2010. It is based on the May 2, 2010 New York Times article 'Driving Shifts Into Reverse'. See http://mbostock.github.io/protovis/ex/driving.html."""
    references = [
        "Fairfield, Hannah. 'Driving Shifts Into Reverse.' New York Times, May 2, 2010"
    ]
    path = "vinyl/data/driving.parquet"


@_return_vinyl_table
class flare(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/flare.parquet"


@_return_vinyl_table
class flare_dependencies(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/flare_dependencies.parquet"


@_return_vinyl_table
class flights_10k(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/flights_10k.parquet"


@_return_vinyl_table
class flights_200k(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/flights_200k.parquet"


@_return_vinyl_table
class flights_20k(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/flights_20k.parquet"


@_return_vinyl_table
class flights_2k(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/flights_2k.parquet"


@_return_vinyl_table
class flights_3m(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/flights_3m.parquet"


@_return_vinyl_table
class flights_5k(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/flights_5k.parquet"


@_return_vinyl_table
class flights_airport(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/flights_airport.parquet"


@_return_vinyl_table
class gapminder(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/gapminder.parquet"


@_return_vinyl_table
class gapminder_health_income(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/gapminder_health_income.parquet"


@_return_vinyl_table
class github(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/github.parquet"


@_return_vinyl_table
class income(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/income.parquet"


@_return_vinyl_table
class iowa_electricity(ExampleDataset):
    description = """The state of Iowa has dramatically increased its production of renewable wind power in recent years. This file contains the annual net generation of electricity in the state by source in thousand megawatthours. The dataset was published in the public domain by the U.S. Energy Information Administration [1]_ and downloaded on May 6, 2018. It is useful for illustrating stacked area charts."""
    references = [
        "https://www.eia.gov/beta/electricity/data/browser/#/topic/0?agg=2,0,1&fuel=vvg&geo=00000g&sec=g&linechart=ELEC.GEN.OTH-IA-99.A~ELEC.GEN.COW-IA-99.A~ELEC.GEN.PEL-IA-99.A~ELEC.GEN.PC-IA-99.A~ELEC.GEN.NG-IA-99.A~~ELEC.GEN.NUC-IA-99.A~ELEC.GEN.HYC-IA-99.A~ELEC.GEN.AOR-IA-99.A~ELEC.GEN.HPS-IA-99.A~&columnchart=ELEC.GEN.ALL-IA-99.A&map=ELEC.GEN.ALL-IA-99.A&freq=A&start=2001&end=2017&ctype=linechart&ltype=pin&tab=overview&maptype=0&rse=0&pin="
    ]
    path = "vinyl/data/iowa_electricity.parquet"


@_return_vinyl_table
class iris(ExampleDataset):
    description = """This classic dataset contains lengths and widths of petals and sepals for 150 iris flowers, drawn from three species. It was introduced by R.A. Fisher in 1936 [1]_."""
    references = [
        "R. A. Fisher (1936). 'The use of multiple measurements in taxonomic problems'. Annals of Eugenics. 7 (2): 179-188."
    ]
    path = "vinyl/data/iris.parquet"


@_return_vinyl_table
class jobs(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/jobs.parquet"


@_return_vinyl_table
class la_riots(ExampleDataset):
    description = """More than 60 people lost their lives amid the looting and fires that ravaged Los Angeles for five days starting on April 29, 1992. This dataset contains metadata about each person, including the geographic coordinates of their death. It was compiled and published by the Los Angeles Times Data Desk [1]_."""
    references = ["http://spreadsheets.latimes.com/la-riots-deaths/"]
    path = "vinyl/data/la_riots.parquet"


@_return_vinyl_table
class londonCentroids(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/londonCentroids.parquet"


@_return_vinyl_table
class lookup_groups(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/lookup_groups.parquet"


@_return_vinyl_table
class lookup_people(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/lookup_people.parquet"


@_return_vinyl_table
class monarchs(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/monarchs.parquet"


@_return_vinyl_table
class normal_2d(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/normal_2d.parquet"


@_return_vinyl_table
class obesity(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/obesity.parquet"


@_return_vinyl_table
class ohlc(ExampleDataset):
    description = """This dataset contains the performance of the Chicago Board Options Exchange `Volatility Index <https://en.wikipedia.org/wiki/VIX>`_ in the summer of 2009."""
    references = None
    path = "vinyl/data/ohlc.parquet"


@_return_vinyl_table
class points(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/points.parquet"


@_return_vinyl_table
class population(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/population.parquet"


@_return_vinyl_table
class population_engineers_hurricanes(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/population_engineers_hurricanes.parquet"


@_return_vinyl_table
class seattle_temps(ExampleDataset):
    description = """This dataset contains hourly temperatures in Seattle during the full year of 2010. The dataset is drawn from public-domain `NOAA data <https://www.weather.gov/disclaimer>`_, and transformed using scripts available at http://github.com/vega/vega_datasets/."""
    references = None
    path = "vinyl/data/seattle_temps.parquet"


@_return_vinyl_table
class seattle_weather(ExampleDataset):
    description = """This dataset contains precipitation totals, temperature extremes, wind speed, and weather type recorded daily in Seattle from 2012 to 2015. The dataset is drawn from public-domain `NOAA data <https://www.weather.gov/disclaimer>`_, and transformed using scripts available at http://github.com/vega/vega_datasets/."""
    references = None
    path = "vinyl/data/seattle_weather.parquet"


@_return_vinyl_table
class sf_temps(ExampleDataset):
    description = """This dataset contains hourly temperatures in San Francisco during the full year of 2010. The dataset is drawn from public-domain `NOAA data <https://www.weather.gov/disclaimer>`_, and transformed using scripts available at http://github.com/vega/vega_datasets/."""
    references = None
    path = "vinyl/data/sf_temps.parquet"


@_return_vinyl_table
class sp500(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/sp500.parquet"


@_return_vinyl_table
class stocks(ExampleDataset):
    description = """Daily closing stock prices for AAPL, AMZN, GOOG, IBM, and MSFT between 2000 and 2010."""
    references = None
    path = "vinyl/data/stocks.parquet"


@_return_vinyl_table
class udistrict(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/udistrict.parquet"


@_return_vinyl_table
class unemployment(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/unemployment.parquet"


@_return_vinyl_table
class unemployment_across_industries(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/unemployment_across_industries.parquet"


@_return_vinyl_table
class uniform_2d(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/uniform_2d.parquet"


@_return_vinyl_table
class us_employment(ExampleDataset):
    description = """In the mid 2000s the global economy was hit by a crippling recession. One result: Massive job losses across the United States. The downturn in employment, and the slow recovery in hiring that followed, was tracked each month by the Current Employment Statistics [1]_ program at the U.S. Bureau of Labor Statistics. This file contains the monthly employment total in a variety of job categories from January 2006 through December 2015. The numbers are seasonally adjusted and reported in thousands. The data were downloaded on Nov. 11, 2018, and reformatted for use in this library. Because it was initially published by the U.S. government, it is in the public domain. Totals are included for the 22 'supersectors' [2]_ tracked by the BLS. The 'nonfarm' total is the category typically used by economists and journalists as a stand-in for the country's employment total. A calculated 'nonfarm_change' column has been appended with the month-to-month change in that supersector's employment. It is useful for illustrating how to make bar charts that report both negative and positive values."""
    references = [
        "https://www.bls.gov/ces/",
        "https://download.bls.gov/pub/time.series/ce/ce.supersector",
    ]
    path = "vinyl/data/us_employment.parquet"


@_return_vinyl_table
class us_state_capitals(ExampleDataset):
    description = """US state capitals with latitude and longitude."""
    references = None
    path = "vinyl/data/us_state_capitals.parquet"


@_return_vinyl_table
class volcano(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/volcano.parquet"


@_return_vinyl_table
class weather(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/weather.parquet"


@_return_vinyl_table
class weball26(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/weball26.parquet"


@_return_vinyl_table
class wheat(ExampleDataset):
    description = """In an 1822 letter to Parliament, William Playfair[1]_, a Scottish engineer who is often credited as the founder of statistical graphics, published an elegant chart on the price of wheat[2]_. It plots 250 years of prices alongside weekly wages and the reigning monarch. He intended to demonstrate that 'never at any former period was wheat so cheap, in proportion to mechanical labour, as it is at the present time.' The electronic dataset was created by Mike Bostock and released into the public domain."""
    references = [
        "https://en.wikipedia.org/wiki/William_Playfair",
        "http://dh101.humanities.ucla.edu/wp-content/uploads/2014/08/Vis_2.jpg",
    ]
    path = "vinyl/data/wheat.parquet"


@_return_vinyl_table
class windvectors(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/windvectors.parquet"


@_return_vinyl_table
class zipcodes(ExampleDataset):
    description = """None"""
    references = None
    path = "vinyl/data/zipcodes.parquet"



# type: ignore
from vinyl import Field, source  # noqa F401
from vinyl import types as t  # noqa F401

from internal_project.resources import supabase_copilot_postgres  # noqa F401


@source(resource=supabase_copilot_postgres)
class Jbin:
    _table = "jbin"
    _unique_name = "supabase_copilot_postgres.postgres.public.Jbin"
    _schema = "public"
    _database = "postgres"
    _twin_path = "data/postgres/public/supabase_copilot_postgres.duckdb"
    _col_replace = {}

    id: t.Int32(nullable=False)



# type: ignore
from vinyl import Field, source  # noqa F401
from vinyl import types as t  # noqa F401

from internal_project.resources import supabase_copilot_postgres  # noqa F401


@source(resource=supabase_copilot_postgres)
class Inferences:
    _table = "inferences"
    _unique_name = "supabase_copilot_postgres.postgres.public.Inferences"
    _schema = "public"
    _database = "postgres"
    _twin_path = "data/postgres/public/supabase_copilot_postgres.duckdb"
    _col_replace = {}

    id: t.UUID(nullable=False)
    created_at: t.Timestamp(timezone="UTC", scale=6, nullable=True)
    input_text: t.String(nullable=True)
    prompt_text: t.String(nullable=True)
    open_ai_prompt_text: t.String(nullable=True)
    model_config: t.JSON(nullable=True)
    prompt_type: t.String(nullable=True)



import os
import sys
from typing import Tuple
import sqlglot

import duckdb

from vinyl.infra.pg_proxy.backends.duckdb import DuckDBConnection
from vinyl.infra.pg_proxy.bv_dialects import BVPostgres, BVDuckDB
from vinyl.infra.pg_proxy.postgres import ProxyServer
from vinyl.infra.pg_proxy.rewrite import Rewriter


class TurntableMetricsRewriter(Rewriter):
    def rewrite(self, sql: str) -> str:
        if sql.lower() == "select pg_catalog.version()":
            return "SELECT 'PostgreSQL 9.3' as version"

        ast = sqlglot.parse_one(sql)
        tables = [t for t in ast.find_all(sqlglot.expressions.Table)]

        metric_ctes = []
        for table in tables:
            if table.db == "metrics":
                # todo: get the real metrics code from ibis here
                cte_sql = "SELECT * FROM source"
                cte_name = f"metrics_{table.name}"
                # cte = sqlglot.parse_one(
                #     f"WITH {cte_name} AS ({cte_sql}) SELECT 1"
                # ).find(sqlglot.expressions.CTE)
                metric_ctes.append((cte_name, cte_sql))

        if len(metric_ctes) > 0:
            for alias, sql in metric_ctes:
                ## todo: clean this up with sqlglot functions to handle prepend
                metrics_sql = f"WITH {alias} AS ( {sql} ) {ast.sql()};"
            return super().rewrite(metrics_sql)

        return super().rewrite(sql)


rewriter = TurntableMetricsRewriter(BVPostgres(), BVDuckDB())


def create(
    db: duckdb.DuckDBPyConnection, host_addr: Tuple[str, int], auth: dict = None
) -> ProxyServer:
    server = ProxyServer(host_addr, DuckDBConnection(db), rewriter=rewriter, auth=auth)
    return server


if __name__ == "__main__":
    if len(sys.argv) < 2:
        print("(Turntable) Using in-memory DuckDB")
        db = duckdb.connect()
    else:
        print("(Turntable) Using DuckDB database at %s" % sys.argv[1])
        db = duckdb.connect(sys.argv[1])

    bv_host = "0.0.0.0"
    bv_port = 5433

    if "HOST" in os.environ:
        bv_host = os.environ["HOST"]

    if "PORT" in os.environ:
        bv_port = int(os.environ["PORT"])

    address = (bv_host, bv_port)
    server = create(db, address)
    ip, port = server.server_address
    print(f"Listening on {ip}:{port}")

    try:
        server.serve_forever()
    finally:
        server.shutdown()
        db.close()



## Needs to happen because nuitka removes all .py files, and in this case, we need to keep them. We zip the folder before the build with this script.
if __name__ == "__main__":
    import shutil

    shutil.make_archive(
        "vinyl/cli/_project_scaffolding", "zip", "vinyl/cli/_project_scaffolding"
    )



from enum import Enum


class FillOptions(Enum):
    null = "null"
    previous = "previous"
    next = "next"


class WindowType(Enum):
    rows = "rows"
    range = "range"


class AssetType(Enum):
    MODEL = "model"
    METRIC = "metric"



import inspect
import os
import signal
from enum import Enum
from functools import wraps
from typing import get_type_hints

from pydantic.validate_call_decorator import validate_call

_validate = validate_call(
    validate_return=False, config={"arbitrary_types_allowed": True}
)


def _with_modified_env(var_name: str, new_value: str):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Save the original value of the environment variable (if it exists)
            original_value = os.environ.get(var_name, None)

            # Set the new value for the environment variable
            os.environ[var_name] = new_value

            try:
                # Call the actual function
                result = func(*args, **kwargs)
            finally:
                # Restore the original value of the environment variable
                if original_value is not None:
                    os.environ[var_name] = original_value
                else:
                    # The environment variable was not set before, so remove it
                    del os.environ[var_name]

            return result

        return wrapper

    return decorator


class ModifiedEnv:
    def __init__(self, var_name: str, new_value: str):
        self.var_name = var_name
        self.new_value = new_value
        self.original_value = None

    def __enter__(self):
        # Save the original value of the environment variable (if it exists)
        self.original_value = os.environ.get(self.var_name, None)

        # Set the new value for the environment variable
        os.environ[self.var_name] = self.new_value

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Restore the original value of the environment variable
        if self.original_value is not None:
            os.environ[self.var_name] = self.original_value
        else:
            # The environment variable was not set before, so remove it
            del os.environ[self.var_name]


def find_enum_arguments(func):
    # Get the signature of the function
    # Iterate over the parameters of the function
    out = {}
    type_hints = get_type_hints(func)
    # Iterate over the type hints of the function
    for name, hint in type_hints.items():
        # Check if the type is a subclass of enum.Enum
        if inspect.isclass(hint) and issubclass(hint, Enum):
            out[name] = hint

    return out


# Define the timeout handler
def timeout_handler(signum, frame):
    raise TimeoutError("Function has timed out")


# Decorator to add a timeout to a function
def timeout(seconds):
    def decorator(func):
        def wrapper(*args, **kwargs):
            # Set the signal handler and the alarm
            signal.signal(signal.SIGALRM, timeout_handler)
            signal.alarm(seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                # Disable the alarm after function executes
                signal.alarm(0)
            return result

        return wrapper

    return decorator



import hashlib
import os
import random
import re
import string
from base64 import urlsafe_b64decode
from urllib.parse import urlparse

from bidict import bidict
from cryptography.fernet import Fernet


def _generate_random_ascii_string(length: int = 30) -> str:
    return "".join(random.choice(string.ascii_lowercase) for i in range(length))


def _split_interval_string(s: str) -> tuple[int, str]:
    for i, char in enumerate(s):
        if not char.isdigit():
            return int(s[:i]), s[i:]
    return int(s), ""


def _extract_uri_scheme(path: str) -> str:
    result = urlparse(path)
    return result.scheme


def _make_python_identifier(str_: str) -> str:
    return re.sub(r"\W|^(?=\d)", "_", str_)


def _create_reproducible_hash(input_string: str) -> str:
    # Encode the input string
    encoded_string = input_string.encode()

    # Create a SHA-32 hash object
    sha256_hash = hashlib.md5()

    # Update the hash object with the bytes-like object (encoded string)
    sha256_hash.update(encoded_string)

    # Generate the hexadecimal representation of the digest
    hex_digest = sha256_hash.hexdigest()

    return hex_digest


def _replace_with_dict(
    text: str, replace_dict: dict[str, str] | bidict[str, str]
) -> str:
    for k, v in replace_dict.items():
        text = text.replace(k, v)
    return text


def _escape_triple_quotes(text):
    if text is None:
        return None
    if "'''" in text:
        text = text.replace("'''", "\\'\\'\\'")
    if '"""' in text:
        text = text.replace('"""', '\\"\\"\\"')
    return text


def decrypt_secret(secret: str) -> str:
    cipher = Fernet(os.getenv("ENCRYPTION_KEY"))

    decrypted_secret = cipher.decrypt(urlsafe_b64decode(secret)).decode("utf-8")
    return decrypted_secret



from __future__ import annotations

import itertools
from functools import lru_cache
from typing import Any, Literal, TypeAlias

import ibis
import ibis.expr.types as ir
import rustworkx as rx

from vinyl.lib.field import Field
from vinyl.lib.graph import VinylGraph
from vinyl.lib.utils.sequence import _find_min_length_list_index

base_join_type: TypeAlias = (
    str
    | ir.BooleanColumn
    | Literal[True]
    | Literal[False]
    | tuple[
        str | ir.Column,
        str | ir.Column,
    ]
)

AUTO_JOIN_DEFAULT_HOW = "left"
MANUAL_JOIN_DEFAULT_HOW = "inner"


class JoinNodeInfo:
    _u_op: ir.Expr | None
    _u_col_name: str
    _v_name: str
    _source_tbl: ir.Expr
    _source_vinyl_table: Any  # will be VinylTable

    def _get_field(self):
        return getattr(self._source_vinyl_table._source_class, self._u_col_name)


class JoinHelper:
    _hash_: str
    _join_index: int
    _node_dict: dict[ir.Expr, ir.Expr]
    _node_info: list[JoinNodeInfo]


@lru_cache(None)
def _get_path(g: rx.PyDiGraph, u: int, v: int) -> rx.PathMapping:
    return rx.dijkstra_shortest_paths(g, source=u, target=v)  # type: ignore[attr-defined]


def _get_routes(
    g: rx.PyDiGraph, tbl_locations: list[int], len_=None
) -> list[list[int] | None]:
    if len is None:
        path_options = [list(p) for p in itertools.permutations(tbl_locations)]
    else:
        path_options = [list(p) for p in itertools.permutations(tbl_locations, len_)]

    # caching inside function so we don't recalucate where we don't need to
    path_routes = []
    for path in path_options:
        path_route_it: list[int] | None = []
        for i, loc in enumerate(path):
            if path_route_it is None:
                continue
            if i < len(path) - 1:
                subpath = _get_path(g, loc, path[i + 1])
                if not subpath:
                    path_route_it = None
                    continue

                subpath_keys = [j for j in [i for i in subpath.values()][0]]
                if len(path_route_it) != 0:
                    path_route_it.pop(-1)
                path_route_it.extend(subpath_keys)
        path_routes.append(path_route_it)

    return path_routes


def _get_path_nodes_and_edges(
    non_trivial_g: rx.PyDiGraph, path_routes: list[list[int] | None]
) -> tuple[list[int], list[Any], list[tuple[str, str] | bool]]:
    # find the shortest path
    min_path_index = _find_min_length_list_index(path_routes)
    if min_path_index is None:
        raise ValueError("No join path found, please join manually")
    final_route = path_routes[min_path_index]
    if final_route is None:
        raise ValueError("No join path found, please join manually")
    final_node_values = [non_trivial_g[i] for i in final_route]
    edge_index_pairs = [i for i in non_trivial_g.edge_list()]
    edge_values = non_trivial_g.edges()
    path_edges = []
    for i, key in enumerate(final_route):
        if i < len(final_route) - 1:
            index = edge_index_pairs.index((key, final_route[i + 1]))
            path_edges.append(edge_values[index])

    return final_route, final_node_values, path_edges


def _auto_join_helper(
    tbls: list[ir.Expr], allow_cross_join: bool = False
) -> tuple[list[ir.Expr], list[tuple[str, str] | bool]]:
    g = Field._relations
    non_trivial_connections: set[int] = set()
    for se in rx.weakly_connected_components(g):  # type: ignore[attr-defined]
        if len(se) > 1:
            non_trivial_connections = non_trivial_connections.union(se)
    non_trivial_g = g.subgraph(list(non_trivial_connections))
    node_hash = {}
    node_dict_store = {}

    # prepare data
    tbl_dict: dict[str, JoinHelper] = {}
    for tbl in tbls:
        hash_ = str(hash(tbl))
        tbl_dict[hash_] = base_it = JoinHelper()
        if hash_ not in node_hash:
            base_it._join_index = node_hash[hash_] = non_trivial_g.add_node(tbl)
        else:
            base_it._join_index = node_hash[hash_]
        node_dict_store[base_it._join_index] = base_it._node_dict = (
            VinylGraph._find_unaltered_cols(tbl)
        )
        base_it._node_info = []
        for u, v in base_it._node_dict.items():
            node_info = JoinNodeInfo()
            node_info._u_op = u.op()
            node_info._u_col_name = u.get_name()
            node_info._source_tbl = node_info._u_op.args[0].to_expr()
            node_info._source_vinyl_table = Field._source_adj_tbl_dict[
                str(hash(node_info._source_tbl))
            ]
            node_info._v_name = v.name
            base_it._node_info.append(node_info)

    non_trivial_nodes = non_trivial_g.nodes()

    # make n = 1 leaps
    for tbl, base_it in tbl_dict.items():
        for base_info in base_it._node_info:
            for i, node in enumerate(non_trivial_nodes):
                # find which source node is relevant
                if hash(node) == hash(base_info._source_vinyl_table):
                    # check which column type it is
                    field = base_info._get_field()
                    if field.primary_key or field.unique:
                        # connect to the source node in both directions
                        non_trivial_g.add_edge(
                            i,
                            base_it._join_index,
                            (base_info._u_col_name, base_info._v_name),
                        )
                        non_trivial_g.add_edge(
                            base_it._join_index,
                            i,
                            (base_info._v_name, base_info._u_col_name),
                        )

    # make n = 2 and n = 3 leaps
    for tbl, base_it in tbl_dict.items():
        for base_info in base_it._node_info:
            for i, node in enumerate(non_trivial_nodes):
                # find which source node is relevant
                if hash(node) == hash(base_info._source_vinyl_table):
                    # make foreign degree connections
                    field = base_info._get_field()
                    if field.foreign_key is not None:
                        in_edges = non_trivial_g.in_edges(i)
                        for edge in in_edges:
                            in_col, out_col = edge[-1]
                            if out_col == base_info._u_col_name:
                                # make second degree connections
                                non_trivial_g.add_edge(
                                    edge[0],
                                    base_it._join_index,
                                    (in_col, base_info._v_name),
                                )
                                edges2 = non_trivial_g.in_edges(edge[0])
                                for edge2 in edges2:
                                    in_col2, out_col2 = edge2[-1]
                                    if out_col2 == in_col:
                                        # make third degree connections
                                        non_trivial_g.add_edge(
                                            edge2[0],
                                            base_it._join_index,
                                            (in_col2, base_info._v_name),
                                        )
                    # make primary key connections
                    if field.primary_key:
                        out_edges = non_trivial_g.out_edges(i)
                        for edge in out_edges:
                            in_col, out_col = edge[-1]
                            if in_col == base_info._u_col_name:
                                # make second degree connections
                                non_trivial_g.add_edge(
                                    base_it._join_index,
                                    edge[1],
                                    (base_info._v_name, out_col),
                                )
                                edges2 = non_trivial_g.out_edges(edge[1])
                                for edge2 in edges2:
                                    in_col2, out_col2 = edge2[-1]
                                    if in_col2 == out_col:
                                        # make third degree connections
                                        non_trivial_g.add_edge(
                                            base_it._join_index,
                                            edge2[1],
                                            (base_info._v_name, out_col2),
                                        )
    # find shortest path
    tbl_locations = [i for i in node_dict_store.keys()]
    path_routes = _get_routes(non_trivial_g, tbl_locations)

    # find the shortest path
    try:
        final_route, final_node_values, path_edges = _get_path_nodes_and_edges(
            non_trivial_g, path_routes
        )
        return final_node_values, path_edges

    except ValueError:
        span = len(tbls)
        if not allow_cross_join:
            raise ValueError(
                "No join path found and cross joins not allowed, please join manually"
            )
        elif span == 2:
            return tbls, [True]  # cross join
        else:
            path_routes_wo1 = _get_routes(non_trivial_g, tbl_locations, len_=span - 1)
            try:
                # up to one cross join is allowed, otherwise error out
                (
                    final_route_wo1,
                    final_node_values_wo1,
                    path_edge_wo1,
                ) = _get_path_nodes_and_edges(non_trivial_g, path_routes_wo1)
                excluded_tbl_num = list(set(tbl_locations) - set(final_route))
                assert len(excluded_tbl_num) == 1
                excluded_tbl = non_trivial_g[excluded_tbl_num[0]]
                return final_node_values_wo1 + [excluded_tbl], path_edge_wo1 + [True]

            except ValueError:
                raise ValueError(
                    "No join path found for more than  tables, please join manually"
                )


def _convert_auto_join_to_ibis_join(
    final_nodes: list[Any],
    edges: list[tuple[str, str] | bool],
    how: Literal[
        "inner",
        "left",
        "outer",
        "right",
        "semi",
        "anti",
        "any_inner",
        "any_left",
        "left_semi",
    ] = AUTO_JOIN_DEFAULT_HOW,  # type: ignore[assignment]
) -> Any:
    # convert to ibis join
    for (
        i,
        node,
    ) in enumerate(final_nodes):
        if i == len(final_nodes) - 1:
            break

        edge: tuple[str, str] | bool = edges[i]

        if isinstance(edge, bool):
            predicate: list[tuple[str, str] | bool] = [edge]
        else:
            predicate = [(edge[0], edge[1])]

        if i == 0:
            join = ibis.join(
                left=node.tbl,
                right=final_nodes[i + 1].tbl,
                how=how,
                predicates=predicate,
            )
        else:
            join = ibis.join(
                left=join.tbl,
                right=final_nodes[i + 1].tbl,
                how=how,
                predicates=predicate,
            )
    return join



import importlib
import itertools
import os
import sys
from pathlib import Path
from types import ModuleType

import toml
from vinyl.lib.utils.pkg import _find_nearest_pyproject_toml_directory


def _get_root() -> Path | None:
    # Create a Path object for the start directory
    path = Path().resolve()

    # Traverse up through the parent directories
    for parent in itertools.chain([path], path.parents):
        pyproject_toml = parent / "pyproject.toml"
        if pyproject_toml.exists():
            return pyproject_toml.parent.absolute()

    return None


class PyProjectSettings:
    _toml: dict

    def __init__(self, path: Path | None = None):
        if path is None:
            path = _get_root()
            if path is None:
                raise FileNotFoundError("pyproject.toml not found")
        with open(path / "pyproject.toml") as f:
            self._toml = toml.load(f)

    def _get_setting(self, key: str) -> str | None:
        return self._toml["tool"]["vinyl"].get(key, None)


def _get_project_module_name() -> str:
    root_path = _find_nearest_pyproject_toml_directory()

    with open(root_path, "r") as file:
        data = toml.load(file)

    return data["tool"]["vinyl"]["module_name"]


def _load_project_module() -> ModuleType:
    # Backup the original sys.modules
    try:
        module_name = _get_project_module_name()
    except KeyError:
        raise ValueError("Can't find a project pyproject.toml file.")

    sys.path.append(os.getcwd())
    imported_module = importlib.import_module(module_name)
    return imported_module



from enum import Enum
from typing import Any, Callable, Literal

from vinyl.lib.column_methods import (
    ColumnBuilder,
    ColumnListBuilder,
    base_column_type,
    column_type_without_dict,
)
from vinyl.lib.settings import PyProjectSettings


class geom(Enum):
    from lets_plot import (
        aes,
        geom_area,
        geom_area_ridges,
        geom_bar,
        geom_bin2d,
        geom_boxplot,
        geom_density,
        geom_histogram,
        geom_line,
        geom_point,
        geom_smooth,
        geom_violin,
        position_dodge,
        ylab,
    )
    from lets_plot.plot.core import LayerSpec

    scatter: Callable[..., LayerSpec] = geom_point()
    line: Callable[..., LayerSpec] = geom_line()
    bar: Callable[..., LayerSpec] = geom_bar(stat="identity", position=position_dodge())
    area: Callable[..., LayerSpec] = geom_area()
    stacked_bar: Callable[..., LayerSpec] = geom_bar(stat="identity")
    percent_bar: Callable[..., LayerSpec] = (
        geom_bar() + aes(y="..prop..") + ylab("Percent of total")
    )
    histogram: Callable[..., LayerSpec] = geom_histogram()
    histogram_2d: Callable[..., LayerSpec] = geom_bin2d()
    violin: Callable[..., LayerSpec] = geom_violin()
    boxplot: Callable[..., LayerSpec] = geom_boxplot()
    density: Callable[..., LayerSpec] = geom_density()
    ridge: Callable[..., LayerSpec] = geom_area_ridges()
    trendline_lm: Callable[..., LayerSpec] = geom_smooth()
    trendline_loess: Callable[..., LayerSpec] = geom_smooth(method="loess")


class BaseChart:
    _mode: Literal["light", "dark"] = "dark"
    _geoms: geom | list[geom]
    _source: Any  # will be VinylTable, but use Any to avoid recursion
    x: base_column_type
    y: base_column_type | None
    color: base_column_type | None
    fill: base_column_type | None
    size: base_column_type | None
    alpha: base_column_type | None
    facet: column_type_without_dict | None
    _coord_flip: bool = False

    def __init__(
        self,
        geoms: geom | list[geom],
        source: Any,  # will be VinylTable, but use Any to avoid recursion
        x: base_column_type,
        y: base_column_type | None = None,
        color: base_column_type | None = None,
        fill: base_column_type | None = None,
        size: base_column_type | None = None,
        alpha: base_column_type | None = None,
        facet: column_type_without_dict | None = None,
        coord_flip: bool = False,
    ):
        self._geoms = geoms
        self._source = source
        self.x = x
        self.y = y
        self.color = color
        self.fill = fill
        self.size = size
        self.alpha = alpha
        self.facet = facet
        self.coord_flip = coord_flip

    def _show(self):
        from lets_plot import (
            aes,
            coord_flip,
            facet_grid,
            facet_wrap,
            flavor_darcula,
            ggplot,
            scale_x_datetime,
            scale_y_datetime,
        )

        adj_facet = self.facet if isinstance(self.facet, list) else [self.facet]
        all_cols = [
            x
            for x in [self.x]
            + [self.y]
            + [self.color]
            + [self.fill]
            + [self.size]
            + [self.alpha]
            + adj_facet
            if x is not None
        ]

        adj_data = self._source.define(all_cols).execute("pandas")

        ## make sure all cols are in there,
        vinyl_x = ColumnBuilder(self._source.tbl, self.x)
        type_x = vinyl_x._type
        if self.y is not None:
            vinyl_y = ColumnBuilder(self._source.tbl, self.y)
            type_y = vinyl_y._type
        aes_dict = {}
        for var in ["x", "y", "color", "fill", "size"]:
            attr = getattr(self, var)
            if attr is not None:
                aes_dict[var] = ColumnBuilder(self._source.tbl, attr)._name

        plot = ggplot(adj_data, aes(**aes_dict))
        if isinstance(self._geoms, list):
            for g in self._geoms:
                plot += g.value
        elif self._geoms is not None:
            plot += self._geoms.value
        if self.facet is not None:
            facet_names = ColumnListBuilder(self._source.tbl, adj_facet)._names
            if len(adj_facet) > 1:
                plot += facet_grid(
                    facet_names[0],
                    facet_names[1],
                )
            else:
                plot += facet_wrap(facet_names[0])
        if type_x.is_timestamp():
            plot += scale_x_datetime()
        if self.y is not None and type_y.is_timestamp():
            plot += scale_y_datetime()
        if self.coord_flip:
            plot += coord_flip()

        if PyProjectSettings()._get_setting("dark-mode") is True:
            plot += flavor_darcula()

        return plot



import toml
import subprocess


def generate_autocomplete_hints(project_names: list = []):
    subprocess.run(
        f"poetry run stubgen --include-docstrings -p {' '.join(project_names)} -o .",
        shell=True,
    )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--project-names", nargs="+")
    args = parser.parse_args()
    generate_autocomplete_hints(args.project_names)



# type: ignore
from vinyl import Field, source  # noqa F401
from vinyl import types as t  # noqa F401

from internal_project.resources import local_filesystem  # noqa F401


@source(resource=local_filesystem)
class Oil:
    _table = "oil"
    _unique_name = "local_filesystem.Oil"
    _path = "data/csvs/oil.csv"
    _row_count = 1218
    _col_replace = {}

    date: t.Date(nullable=True)
    dcoilwtico: t.Float64(nullable=True)



from vega_datasets import data

if __name__ == "__main__":
    examples_class_path = "vinyl/examples.py"

    with open(examples_class_path, "w") as f:
        f.write("""
import os

import ibis

from vinyl.lib.table import VinylTable


def _return_vinyl_table(cls):
    def wrapper(*args, **kwargs) -> VinylTable:
        adj_path_base = os.path.dirname(os.path.dirname(__file__))
        return VinylTable(
            ibis.read_parquet(
                os.path.join(adj_path_base, cls.path), table_name=cls.__name__
            )._arg
        )
            
    wrapper.description = cls.description
    wrapper.references = cls.references

    return wrapper

class ExampleDataset:
    description: str
    references: str
    path: str

""")

    for dataset in dir(data):
        obj = getattr(data, dataset)
        path = f"vinyl/data/{dataset}.parquet"
        try:
            df = obj()
            df.to_parquet(path)
        except Exception:
            # file is not a dataframe
            continue

        if dataset[0].isdigit():
            dataset = f"_{dataset}"
        with open(examples_class_path, "a") as f:
            f.write(f"""
@_return_vinyl_table
class {dataset}(ExampleDataset):
    description = '''{obj.description}'''
    references = {obj.references}
    path = "vinyl/data/{dataset}.parquet"
    """)



from typing import Any, Callable, TypeVar

import sqlglot
import sqlglot.expressions as exp

DecoratedCallable = TypeVar("DecoratedCallable", bound=Callable[..., Any])


class Rewriter:
    def __init__(self, read: sqlglot.Dialect, write: sqlglot.Dialect):
        self._relations = {}
        self._read = read
        self._write = write

    def relation(self, name: str) -> Callable[[DecoratedCallable], DecoratedCallable]:
        def decorator(func: DecoratedCallable) -> DecoratedCallable:
            self._relations[name] = func
            return func

        return decorator

    def rewrite(self, sql: str) -> str:
        try:
            stmts = self._read.parse(sql)
            ret = []
            for stmt in stmts:
                ret.append(self.rewrite_one(stmt))
            return ";\n".join(self._write.generate(s) for s in ret)
        except Exception as e:
            # TODO: log this
            print(e)
            print(f"failed to rewrite: {sql}")
            return sql

    def rewrite_one(self, expression: exp.Expression) -> exp.Expression:
        def _expand(node: exp.Expression):
            if isinstance(node, exp.Table):
                name = exp.table_name(node)
                if name in self._relations:
                    source = self._relations[name]
                    subquery = exp.paren(exp.maybe_parse(source()))
                    if node.alias:
                        subquery = exp.alias_(subquery, node.alias)
                    subquery.comments = [f"source: {name}"]
                    return subquery
            return node

        return expression.transform(_expand, copy=True)


if __name__ == "__main__":
    rewriter = Rewriter(sqlglot.dialects.Presto(), sqlglot.dialects.DuckDB())

    @rewriter.relation("schema.test")
    def test():
        return "SELECT 1 as a, 'foo' as b"

    print(rewriter.rewrite("SELECT * FROM schema.test"))



import pytest

from vinyl.lib.connect import DatabaseFileConnector, FileConnector


def test_local_no_args():
    with pytest.raises(TypeError):
        FileConnector(path=None)._connect()


def test_connect_with_duckdb_file():
    conn = DatabaseFileConnector(
        path="../tests/vinyl/fixtures/test.duckdb", tables=["*.*.*"]
    )._connect()

    assert "taxi_sample" in conn.list_tables(database=("test", "main"))
    assert (
        conn.table(database=("test", "main"), name="taxi_sample").count().execute()
        == 100000
    )


def test_connect_with_filepath():
    conn = FileConnector(path="../tests/vinyl/fixtures/data/iris.parquet")._connect()

    assert conn.table("iris").count().execute() == 150


def test_connect_with_dir():
    conn = FileConnector(path="../tests/vinyl/fixtures/data")._connect()
    print(conn.list_tables())
    assert set(["bitcoin", "iris"]).issubset(set(conn.list_tables()))



from vinyl.lib.project import Project


def test_amount_base():
    project = Project.bootstrap()
    table = project._get_model("amount_base")
    assert "total_amount" in table.columns
    assert len(table.execute(limit=None)) == 100000


def test_amount_by_trip_distance():
    project = Project.bootstrap()
    table = project._get_model("amount_by_trip_distance")
    assert table.columns == ["trip_distance", "total_amount"]
    assert len(table.execute(limit=None)) == 100000


def test_amount_metric_by_day():
    project = Project.bootstrap()
    table = project._get_model("amount_metric_by_day")

    assert "total_fare" in table.columns
    assert len(table.execute(limit=None)) > 0


def test_fare_ratio_by_month():
    project = Project.bootstrap()
    table = project._get_model("fare_ratio_by_month")

    assert "fare_ratio" in table.columns
    assert len(table.columns) == 4
    assert len(table.execute(limit=None)) > 0


def test_join():
    project = Project.bootstrap()
    table = project._get_model("store_txns")
    assert len(table.execute(limit=None)) == 83488



from typing import Any


def _find_min_length_list_index(list_of_lists: list[list[Any] | None]) -> int | None:
    # Check if the list is empty
    if not list_of_lists:
        return None  # Return None if the list is empty

    # Iterate through the list of lists
    not_started = True
    min_index = None

    for index, lst in enumerate(list_of_lists):
        if lst is None:
            continue
        elif not_started:
            min_length = len(lst)
            min_index = index
            not_started = False
        # Update minimum length and index if a shorter list is found
        elif len(lst) < min_length:
            min_length = len(lst)
            min_index = index

    return min_index



from __future__ import annotations

import os
from dataclasses import dataclass
from importlib import import_module
from typing import Any

from vinyl.lib.settings import _get_project_module_name, _load_project_module
from vinyl.lib.utils.functions import _with_modified_env
from vinyl.lib.utils.graph import DAG
from vinyl.lib.utils.pkg import (
    _find_submodules_names,
)


def _load_project_defs() -> Defs:
    imported_module = _load_project_module()
    return imported_module.defs


def _create_dag_from_model_dict_list(models: list[dict]) -> DAG:
    model_dag = DAG()
    for model_dict in models:
        model_dag.add_node(model_dict["unique_name"])
        for dep in model_dict["config"].get("deps", {}).values():
            model_dag.add_edge(dep, model_dict["unique_name"])
    return model_dag


@dataclass
class Defs:
    resources: list[Any]
    sources: list[Any]
    models: list[Any]
    metrics: list[Any]
    graph: DAG | None
    default_resource: Any

    def __init__(
        self,
        resources=[],
        sources=[],
        models=[],
        metrics=[],
        graph=None,
        default_resource=None,
    ):
        self.resources = resources
        self.sources = sources
        self.models = models
        self.metrics = metrics
        self.graph = DAG() if graph is None else graph
        self.db_graph = DAG()
        self.default_resource = default_resource

    def _process(self, asset):
        metric_nodes = set()
        for key in ["resource", "source", "model", "metric"]:
            plural_key = key + "s"
            if hasattr(asset, f"_is_vinyl_{key}"):
                current = getattr(self, plural_key)
                if asset not in current:
                    setattr(
                        self,
                        plural_key,
                        getattr(self, plural_key) + [asset],
                    )
                if key in ["model", "metric"]:
                    self.graph.add_node(asset)
                    for dep in asset._deps:
                        # if isinstance(dep, VinylTable):
                        #     # use original source class
                        #     dep = dep._source_class
                        self.graph.add_edge(dep, asset)
                    if key == "metric":
                        metric_nodes.add(asset)
                if key in ["resource"]:
                    if asset._default:
                        self.default_resource = asset

    def _post_process(self):
        # remove and reconnect metric nodes for db_graph, since they don't produce actual tables
        self.db_graph = self.graph.copy()
        self.db_graph.remove_nodes_and_reconnect(self.metrics)

    @classmethod
    def load(cls, proj_name: str | None = None, parallel: bool = False) -> Defs:
        def fn(cls):
            defs_instance = Defs()
            if proj_name is not None:
                project_module_name = proj_name
            else:
                project_module_name = _get_project_module_name()
            modules = []
            modules.append(import_module(f"{project_module_name}.resources"))
            modules.append(import_module(f"{project_module_name}.sources"))
            no_models_vinyl = os.getenv("NO_MODELS_VINYL")
            if no_models_vinyl is None or no_models_vinyl == "False":
                modules.append(import_module(f"{project_module_name}.models"))

            for m in modules:
                for name in _find_submodules_names(m):
                    module = import_module(name)
                    for func in dir(module):
                        attr = getattr(module, func)
                        if func != "_":
                            # order here is important, resources, but be imported before sources
                            defs_instance._process(attr)

            defs_instance._post_process()
            return defs_instance

        if parallel:
            # make duckdb read only
            fn = _with_modified_env("DUCKDB_READ_ONLY", "1")(fn)

        return fn(cls)



import functools
import glob
import hashlib
import os
import subprocess
import tempfile
from contextlib import contextmanager
from functools import wraps
from typing import Any, Callable

import orjson


def _create_dirs_with_init_py(end_path: str, start_dir: str = ".") -> None:
    # Ensure the start_dir is an absolute path
    start_dir = os.path.abspath(start_dir)

    # Construct the full path to the target directory
    # Split the relative path into its components
    relative_path = os.path.relpath(end_path, start_dir)
    dirs = relative_path.split(os.sep)

    # Create __init__.py along the relative path
    for i in range(1, len(dirs) + 1):
        # Construct the path up to the current depth within the relative path
        current_dir = os.path.join(start_dir, os.sep.join(dirs[:i]))
        if not os.path.exists(current_dir):
            # Create the directory if it doesn't exist
            os.makedirs(current_dir, exist_ok=True)
        init_file = os.path.join(current_dir, "__init__.py")
        if not os.path.isfile(init_file):
            # Create __init__.py if it doesn't exist
            open(init_file, "a").close()

    return None


def _file_hash(file_path: str, hash_func: Callable[..., Any] = hashlib.md5) -> str:
    hash_obj = hash_func()
    with open(file_path, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_obj.update(chunk)
    return hash_obj.hexdigest()


def _get_directory_hashes(
    directory: str, hash_func: Callable[..., Any] = hashlib.md5
) -> dict[str, str]:
    hashes = {}
    for root, _, files in os.walk(directory):
        for filename in files:
            file_path = os.path.join(root, filename)
            if os.path.isfile(file_path):
                relative_path = os.path.relpath(file_path, directory)
                hashes[relative_path] = _file_hash(file_path, hash_func)
    return hashes


def _get_changed_files(directory: str, hash_func: Callable[..., Any] = hashlib.md5):
    def decorator_check_files(func):
        @functools.wraps(func)
        def wrapper_check_files(*args, **kwargs):
            # Check file hashes before function execution
            before_hashes = _get_directory_hashes(directory, hash_func)
            func(*args, **kwargs)  # Execute the function
            # Check file hashes after function execution
            after_hashes = _get_directory_hashes(directory, hash_func)

            if before_hashes != after_hashes:
                out = {}
                for file_path in before_hashes:
                    if before_hashes[file_path] != after_hashes.get(file_path):
                        out[file_path] = (
                            before_hashes[file_path],
                            after_hashes.get(file_path),
                        )
                return out

            return {}

        return wrapper_check_files

    return decorator_check_files


def _preserve_file_state(file_path: str):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Read the original content of the file
            if os.path.exists(file_path):
                with open(file_path, "r") as file:
                    original_content = file.read()
            else:
                original_content = None

            # Execute the function
            result = func(*args, **kwargs)

            # Check if the file content has changed
            if os.path.exists(file_path):
                with open(file_path, "r") as file:
                    new_content = file.read()
                if new_content != original_content:
                    # If changed, revert to original content
                    with open(file_path, "w") as file:
                        file.write(original_content)

            return result

        return wrapper

    return decorator


def preserve_file_state_on_failure(file_path: str, overwrite=True, bytes=False):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Read the original content of the file
            if os.path.exists(file_path):
                with open(file_path, "r" if not bytes else "rb") as file:
                    original_content = file.read()
                if overwrite:
                    os.remove(file_path)
            else:
                original_content = None

            try:
                # Execute the function
                result = func(*args, **kwargs)

            except Exception as e:
                # Check if the file content has changed
                if os.path.exists(file_path):
                    with open(file_path, "r" if not bytes else "rb") as file:
                        new_content = file.read()
                    if new_content != original_content:
                        # If changed, revert to original content
                        with open(file_path, "w" if not bytes else "wb") as file:
                            file.write(original_content)

                raise e

            return result

        return wrapper

    return decorator


@contextmanager
def cd(newdir):
    prevdir = os.getcwd()
    os.chdir(newdir)
    try:
        yield
    finally:
        os.chdir(prevdir)


def load_orjson(path):
    with open(path, "r") as f:
        contents = f.read()
    out = orjson.loads(contents)
    return out


def save_orjson(path, contents):
    to_save = orjson.dumps(contents).decode("utf-8")
    with open(path, "w") as f:
        f.write(to_save)


def ruff_format_text(content: str) -> str:
    with tempfile.NamedTemporaryFile(mode="r+", delete=True, suffix=".py") as tf:
        tf.write(content)
        tf.flush()
        subprocess.run(
            ["ruff", "format", tf.name], check=True, stdout=subprocess.DEVNULL
        )  # suppress ruff output
        with open(tf.name, "r") as f:
            formatted_content = f.read()
    return formatted_content


def adjust_path(path: str) -> str:
    return os.path.abspath(os.path.expanduser(path))


def file_exists_in_directory(file_name, directory_path):
    file_path = os.path.join(directory_path, "**", file_name)
    files = glob.glob(file_path, recursive=True)
    if len(files) > 0:
        return files[0]
    else:
        return None



from functools import reduce, wraps
from typing import Literal

import ibis
import ibis.expr.operations as ops
import ibis.expr.types as ir

from vinyl.lib.column import _demote_args
from vinyl.lib.set_methods import (
    AUTO_JOIN_DEFAULT_HOW,
    MANUAL_JOIN_DEFAULT_HOW,
    _auto_join_helper,
    _convert_auto_join_to_ibis_join,
    base_join_type,
)
from vinyl.lib.table import VinylTable, _base_ensure_output
from vinyl.lib.table_methods import _difference_two, _intersect_two, _union_two
from vinyl.lib.utils.functions import _validate


def _ensure_output_left(func):
    @wraps(func)
    def wrapper(left: VinylTable, *args, **kwargs) -> VinylTable:
        return _base_ensure_output(func, left, *args, **kwargs)

    return wrapper


def _ensure_output_first(func):
    @wraps(func)
    def wrapper(first: VinylTable, *args, **kwargs) -> VinylTable:
        return _base_ensure_output(func, first, *args, **kwargs)

    return wrapper


@_demote_args
@_validate
@_ensure_output_left
def join(
    left: VinylTable,
    right: VinylTable,  # will be a VinylTable, but pydantic can't validate
    *args,
    auto: bool = True,
    auto_allow_cross_join: bool = False,
    on: base_join_type | list[base_join_type] = [],
    how: Literal[
        "inner",
        "left",
        "outer",
        "right",
        "semi",
        "anti",
        "any_inner",
        "any_left",
        "left_semi",
    ]
    | None = None,
    lname: str = "",
    rname: str = "{name}_right",
    _how_override: bool = False,  # helper to determine if user has explicitly given how
) -> VinylTable:
    """
    Joins two or more tables together. if `on` is not provided and `auto` is True, the function will attempt to automatically join the tables (including multi-hop joins) based on the relationships defined in the vinyl field graph. If `on` is provided, the function will join the tables based on the provided predicates.

    If `how` is not provided, the function will default to a left join for successful auto joins and an inner join for manual joins. If `how` is provided, the function will use the specified join type. If the auto join fails, the function will raise an error unless `allow_cross_join` is set to True, in which case it will return the cross join.

    By default, all columns from the left and right tables will be included in the output. If you want to exclude columns from the right table, you can use the `select` method to select only the columns you want to keep. If there are duplicate column names in the left and right tables, lname and rname can be used to specify the suffixes to add to the column names from the left and right tables, respectively.
    """
    all_tbs = [left, right, *args]
    adj_all_tbls = []
    if not isinstance(on, list):
        on = [on]

    for tbl in all_tbs:
        if isinstance(tbl.tbl.op(), ops.UnboundTable):
            to_add = tbl._copy(
                mutable=False
            ).select(
                tbl.columns
            )  # handles issues with column replacements when joining directly to an unbound table
        else:
            to_add = tbl._copy(mutable=False)
        adj_all_tbls.append(to_add)

    if auto and on == []:
        nodes, edges = _auto_join_helper(
            adj_all_tbls, allow_cross_join=auto_allow_cross_join
        )
        first = nodes[0]
        other_nodes = nodes[1:]

        # need to ensure output to pass on all nodes into connection dicts, may be more than just left and right. We do this by making all nodes a top-level arg and using the features of ensure_output
        @_ensure_output_first
        def auto_join_iteration(first, *other_nodes, edges):
            nodes = [first, *other_nodes]
            return _convert_auto_join_to_ibis_join(
                nodes,
                edges,
                how=AUTO_JOIN_DEFAULT_HOW
                if how is None
                else how,  # uses left join unless user specifies otherwise, ignoring default
            )

        return auto_join_iteration(
            first,
            *other_nodes,
            edges=edges,
        )

    # adjust left, right, and on to ensure internal consistency
    new_left = adj_all_tbls[0].tbl
    new_right = adj_all_tbls[1].tbl
    new_on = []
    for o in on:
        if isinstance(o, ir.Expr):
            o_it = (
                o.op()
                .replace({left.tbl.op(): new_left.op(), right.tbl.op(): new_right.op()})
                .to_expr()
            )
        else:
            o_it = o
        new_on.append(o_it)

    base_join = ibis.join(
        left=new_left,
        right=new_right,
        predicates=new_on,
        how=MANUAL_JOIN_DEFAULT_HOW if how is None else how,
        lname=lname,
        rname=rname,
    )
    return base_join


@_demote_args
@_validate
@_ensure_output_first
def union(first: VinylTable, *rest: VinylTable, distinct: bool = False) -> VinylTable:
    """
    Compute the set union of multiple table expressions.

    Unlike the SQL UNION operator, this function allows for the union of tables with different schemas. If a column is present in one table but not in another, the column will be added to the other table with NULL values.

    If `distinct` is True, the result will contain only distinct rows. If `distinct` is False, the result will contain all rows from all tables, including duplicates.
    """
    if len(rest) == 0:
        return first

    return reduce(lambda x, y: _union_two(x.tbl, y.tbl, distinct), rest, first)


@_demote_args
@_validate
@_ensure_output_first
def difference(
    first: VinylTable, *rest: VinylTable, distinct: bool = False
) -> VinylTable:
    """
    Compute the set difference of multiple table expressions.

    Unlike the SQL EXCEPT operator, this function allows for the difference of tables with different schemas. If a column is present in one table but not in another, the column will be added to the other table with NULL values.

    If `distinct` is True, the result will contain only distinct rows. If `distinct` is False, the result will contain all rows from the first table, including duplicates.
    """
    if len(rest) == 0:
        return first

    return reduce(lambda x, y: _difference_two(x.tbl, y.tbl, distinct), rest, first)


@_demote_args
@_validate
@_ensure_output_first
def intersect(
    first: VinylTable, *rest: VinylTable, distinct: bool = True
) -> VinylTable:
    """
    Compute the set intersection of multiple table expressions.

    Unlike the SQL INTERSECT operator, this function allows for the intersection of tables with different schemas. If a column is present in one table but not in another, the column will be added to the other table with NULL values.

    If `distinct` is True, the result will contain only distinct rows. If `distinct` is False, the result will contain all rows that are present in all tables, including duplicates.
    """
    if len(rest) == 0:
        return first

    return reduce(lambda x, y: _intersect_two(x.tbl, y.tbl, distinct), rest, first)



from typing import Any

import ibis
from ibis.expr.types import BooleanValue, Column, IntegerColumn, IntegerValue, Value

from vinyl.lib.column import _demote_args_outside_class, _promote_output_outside_class
from vinyl.lib.utils.functions import _validate


def case(
    pairs: list[tuple[BooleanValue, Value]], default: Value | None = None
) -> Value:
    """
    Returns the first value for which the corresponding condition is true. If no conditions are true, return the default.

    Conditions should be specified as a list of tuples, where the first element of each tuple is a boolean expression and the second element is the value to return if the condition is true.
    """
    if len(pairs) == 0:
        raise ValueError("At least one pair must be provided")
    elif len(pairs) == 1:
        return ibis.ifelse(pairs[0][0], pairs[0][1], default)

    out = ibis.case()
    for pair in pairs:
        out = out.when(pair[0], pair[1])
    out = out.else_(default)
    return out.end()


def if_else(condition: Any, true_value: Any | None, false_value: Any | None) -> Value:
    """
    Constructs a conditional expression. If the condition is true, return the true_value; otherwise, return the false_value.

    Can be chained together by making the true_value or false_value another if_else expression.
    """
    return ibis.ifelse(condition, true_value, false_value)


def coalesce(*exprs) -> Value:
    """
    Return the first non-null value in the expression list.
    """
    return ibis.coalesce(*exprs)


def least(*args: Any) -> Value:
    """
    Return the smallest value among the supplied arguments.
    """
    return ibis.least(*args)


def greatest(*args: Any) -> Value:
    """
    Return the largest value among the supplied arguments.
    """
    return ibis.greatest(*args)


def row_number() -> IntegerColumn:
    """
    Returns the current row number.

    This function is normalized across backends to start from 0.
    """
    return ibis.row_number()


def rank(dense: bool = False) -> IntegerColumn:
    """
    Compute position of first element within each equal-value group in sorted order.

    If `dense` don't skip records after ties. See [here](https://learnsql.com/cookbook/whats-the-difference-between-rank-and-dense_rank-in-sql/) for a good primer on the difference.

    """
    if dense:
        return ibis.dense_rank()
    return ibis.rank()


def percent_rank() -> Column:
    """
    Compute the relative rank of a value within a group of values.
    """
    return ibis.percent_rank()


def ntile(n: int | IntegerValue) -> IntegerColumn:
    """
    Divide the rows into `n` buckets, assigning a bucket number to each row.
    """
    return ibis.ntile(n)


# Adjust the function arguments to work with VinylColumn syntax and add validation
case = _validate(_promote_output_outside_class(_demote_args_outside_class(case)))
if_else = _validate(_promote_output_outside_class(_demote_args_outside_class(if_else)))
coalesce = _validate(
    _promote_output_outside_class(_demote_args_outside_class(coalesce))
)
least = _validate(_promote_output_outside_class(_demote_args_outside_class(least)))
greatest = _validate(
    _promote_output_outside_class(_demote_args_outside_class(greatest))
)
row_number = _validate(
    _promote_output_outside_class(_demote_args_outside_class(row_number))
)
rank = _validate(_promote_output_outside_class(_demote_args_outside_class(rank)))
percent_rank = _validate(
    _promote_output_outside_class(_demote_args_outside_class(percent_rank))
)
ntile = _validate(_promote_output_outside_class(_demote_args_outside_class(ntile)))



# type: ignore
from vinyl import Field, source  # noqa F401
from vinyl import types as t  # noqa F401

from internal_project.resources import dbt_bigquery  # noqa F401


@source(resource=dbt_bigquery)
class Companies:
    _table = "companies"
    _unique_name = "dbt_bigquery.analytics-dev-372514.dbt_itracey.Companies"
    _schema = "dbt_itracey"
    _database = "analytics-dev-372514"
    _twin_path = "data/analytics-dev-372514/dbt_itracey/dbt_bigquery.duckdb"
    _col_replace = {}

    domain: t.String(nullable=True)
    last_seen: t.Date(nullable=True)
    num_users: t.Int64(nullable=True)



import toml


def set_version(version: str):
    with open("pyproject.toml", "r") as f:
        pyproject = toml.load(f)
    pyproject["tool"]["cibuildwheel"]["build"] = f"cp{version.replace('.', '')}-*"
    with open("pyproject.toml", "w") as f:
        toml.dump(pyproject, f)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--version")
    args = parser.parse_args()
    set_version(args.version)



# type: ignore
from vinyl import Field, source  # noqa F401
from vinyl import types as t  # noqa F401

from internal_project.resources import local_filesystem  # noqa F401


@source(resource=local_filesystem)
class SampleSubmission:
    _table = "sample_submission"
    _unique_name = "local_filesystem.SampleSubmission"
    _path = "data/csvs/sample_submission.csv"
    _row_count = 28512
    _col_replace = {}

    id: t.Int64(nullable=True)
    sales: t.Float64(nullable=True)



from __future__ import annotations

import copy
from typing import Any, Callable, TypeAlias

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis import _
from ibis import selectors as s
from ibis.common.deferred import Deferred

from vinyl.lib.graph import VinylGraph

base_column_type: TypeAlias = str | ir.Value | Callable[..., Any]

base_boolean_column_type: TypeAlias = ir.BooleanValue | Callable[..., Any]
boolean_column_type: TypeAlias = (
    base_boolean_column_type | list[base_boolean_column_type]
)


column_type_without_dict: TypeAlias = (
    base_column_type | s.Selector | list[base_column_type | s.Selector]
)

column_type: TypeAlias = column_type_without_dict | dict[str, base_column_type]

column_type_all: TypeAlias = column_type_without_dict | list[column_type_without_dict]


# @_validate
def _append_name(col: ir.Value | Deferred, name: str | None = None) -> base_column_type:
    if name is None:
        return col
    try:  # check if v can be named, if it can't, wrap in coalesce
        col_named = col.name(name)
    except Exception:
        col_named = ibis.coalesce(col).name(name)

    return col_named


# @_validate
def _process_cols_arg(
    tbl: ir.Table,
    cols: column_type | None,
    names: list[str | None] | str | None = None,
    passthrough_deferred=False,
) -> list[ir.Value]:
    out: list[ir.Value] = []
    name = names[0] if isinstance(names, list) else names
    if cols is None:
        return [None]

    elif isinstance(cols, ir.Value):
        out = [_append_name(cols, name)]

    elif isinstance(cols, str):
        out = [_append_name(getattr(tbl, cols), name)]

    elif isinstance(cols, Deferred):
        if passthrough_deferred:
            out = [_append_name(cols, name)]
        else:
            out = [_append_name(cols.resolve(tbl), name)]

    elif callable(cols):
        if passthrough_deferred:
            out = [lambda t: _append_name(cols(t), name)]
        else:
            out = [_append_name(cols(tbl), name)]

    elif isinstance(cols, s.Selector):
        out = cols.expand(tbl)

    elif isinstance(cols, list):
        for col in cols:
            out.extend(
                _process_cols_arg(tbl, col, passthrough_deferred=passthrough_deferred)
            )

    elif isinstance(cols, dict):
        for name, col in cols.items():
            out.extend(
                _process_cols_arg(
                    tbl, col, name, passthrough_deferred=passthrough_deferred
                )
            )

    return out


class ColumnBuilder:
    _tbl: ir.Table
    _col: ir.Value | None

    def __init__(
        self,
        tbl: ir.Table,
        col: base_column_type | s.Selector | None,
        passthrough_deferred=False,
    ):
        self._tbl = tbl
        if col is None:
            self._col = col
        else:
            self._col = _process_cols_arg(
                self._tbl, col, passthrough_deferred=passthrough_deferred
            )[0]

    @property
    def _name(self) -> str | None:
        if self._col is None:
            return None

        elif isinstance(self._col, str):
            return self._col

        elif isinstance(self._col, Deferred):
            return self._col.resolve(self._tbl).get_name()

        elif callable(self._col):
            return self._col(self._tbl).get_name()

        return self._col.get_name()

    @property
    def _type(self) -> dt.DataType | None:
        if self._col is None:
            return None

        elif isinstance(self._col, str):
            return getattr(self._tbl, self._col).resolve(self._tbl).type()

        elif isinstance(self._col, Deferred):
            return self._col.resolve(self._tbl).type()

        elif callable(self._col):
            return self._col(self._tbl).type()

        return self._col.type()

    @property
    def _name_as_deferred(self) -> Deferred | None:
        if self._col is None or self._name is None:
            return None
        return getattr(_, self._name)

    def _name_as_deferred_resolved(self, tbl) -> ir.Value | None:
        if self._col is None or self._name is None:
            return None
        col = getattr(tbl, self._name)
        if hasattr(col, "_col"):
            # is VinylColumn
            return col._col
        else:
            # is ibis column
            return col

    @property
    def _lambdaized(self) -> Callable[[ir.Table], ir.Value]:
        if isinstance(self._col, Deferred):
            return self._col.resolve(self._tbl)

        if self._col is None or self._tbl is None:
            return lambda t: None

        if callable(self._col):
            return self._col

        def out(t: ir.Table) -> ir.Value:
            if self._col is None:
                return None
            return self._col.op().replace({self._tbl.op(): t.op()}).to_expr()

        return out

    @property
    def _sources(self) -> list[str] | None:
        if self._col is None:
            return None

        elif isinstance(self._col, Deferred):
            return self._col.resolve(self._tbl)

        elif callable(self._col):
            return self._col(self._tbl)

        return [i.name for i in self._col.op().find_topmost(ops.Field)]

    @property
    def _nodes(self) -> list[ir.Expr]:
        graph = VinylGraph._new_init_from_expr(self._tbl.select(self._col))
        return graph.nodes()

    @property
    def _is_unaltered(self) -> bool:
        allowed = (ops.Field, ops.Relation)
        return all([isinstance(n, allowed) for n in self._nodes])

    @property
    def _is_only_aliased(self) -> bool:
        allowed = (ops.Field, ops.Relation, ops.Alias)
        return all([isinstance(n, allowed) for n in self._nodes])


class ColumnListBuilder:
    _tbl: ir.Table
    _unique: bool
    _cols: list[ColumnBuilder]
    _passthrough_deferred: bool

    def __init__(
        self,
        tbl: ir.Table,
        cols: dict[str, Deferred | ir.Value]
        | list[Deferred | ir.Value | s.Selector]
        | Deferred
        | ir.Value
        | s.Selector
        | None,
        unique: bool = False,
        passthrough_deferred: bool = False,
    ):
        self._tbl = tbl
        self._unique = unique
        self._passthrough_deferred = passthrough_deferred
        self._cols = []

        cols = _process_cols_arg(
            self._tbl, cols, passthrough_deferred=passthrough_deferred
        )
        if cols is None:
            pass
        elif isinstance(cols, list):
            for col in cols:
                if isinstance(col, s.Selector):
                    self._cols.extend(
                        [
                            ColumnBuilder(
                                tbl, c, passthrough_deferred=passthrough_deferred
                            )
                            for c in col.expand(self._tbl)
                        ]
                    )
                else:
                    self._cols.append(
                        ColumnBuilder(
                            tbl, col, passthrough_deferred=passthrough_deferred
                        )
                    )
        else:
            if isinstance(cols, s.Selector):
                self._cols = [
                    ColumnBuilder(tbl, c, passthrough_deferred=passthrough_deferred)
                    for c in cols.expand(self._tbl)
                ]
            else:
                self._cols = [
                    ColumnBuilder(tbl, cols, passthrough_deferred=passthrough_deferred)
                ]

        if unique:
            self._make_unique()

    def __iter__(self):
        return iter(self._cols)

    def __add__(self, other) -> ColumnListBuilder:
        if isinstance(other, ColumnBuilder):
            if self._tbl != other._tbl:
                raise ValueError(
                    "Can only add a VinylColumn to a VinylColumnList from the same VinylTable"
                )
            self._cols.append(other)
            return ColumnListBuilder(
                self._tbl,
                self._cols,
                unique=self._unique,
                passthrough_deferred=self._passthrough_deferred,
            )
        elif not isinstance(other, ColumnListBuilder):
            raise ValueError(
                f"Can only add a VinylColumnList to another VinylColumnList, not a {type(other)}"
            )

        elif self._tbl.__hash__() != other._tbl.__hash__():
            raise ValueError(
                "Can only add a VinylColumnList to another VinylColumnList from the same VinylTable"
            )

        return ColumnListBuilder(
            self._tbl,
            [c._col for c in self._cols] + [c._col for c in other._cols],
            unique=self._unique and other._unique,
            passthrough_deferred=self._passthrough_deferred,
        )

    def __radd__(self, other) -> ColumnListBuilder:
        if isinstance(other, ColumnListBuilder):
            return other.__add__(self)

        return self.__add__(other)

    def _windowize(self, window_, adj_object=False) -> ColumnListBuilder:
        # ensure col can be windowed
        new_obj = self if adj_object else copy.deepcopy(self)
        for col in new_obj._cols:
            if col._col is None:
                continue
            try:
                windowed = col._col.over(window_)
                if isinstance(windowed, Deferred):
                    windowed = windowed.resolve(self._tbl)
                col._col = windowed
            except com.IbisTypeError:
                pass

        return new_obj

    def _make_unique(self) -> None:
        names = []
        unique_cols = []
        zipped = list(zip(self._names, self._queryable))[
            ::-1
        ]  # reverse so latest iteration of column is kept
        for name, col in zipped:
            if name not in names:
                names.append(name)
                unique_cols.append(col)

        unique_cols = unique_cols[::-1]
        self._cols = [
            ColumnBuilder(self._tbl, col, self._passthrough_deferred)
            for col in unique_cols
        ]
        self._unique = True

    def _reset_tbl(self, tbl: ir.Table) -> None:
        self._tbl = tbl

    @property
    def _queryable(self) -> list[ir.Value]:
        out = []
        for col in self._cols:
            if col._col is not None:
                out.append(col._col)
        return out

    @property
    def _lambdaized(self) -> list[Callable[[ir.Table], ir.Value]]:
        return [col._lambdaized for col in self._cols if col._col is not None]

    @property
    def _sources_as_deferred(self) -> list[Deferred]:
        return [getattr(_, col) for col in self._get_direct_col_sources(unique=True)]

    def _sources_as_deferred_resolved(self, tbl) -> list[ir.Value]:
        return [getattr(tbl, col) for col in self._get_direct_col_sources(unique=True)]

    @property
    def _names(self) -> list[str | None]:
        return [col._name for col in self._cols]

    @property
    def _types(self) -> list[dt.DataType | None]:
        return [col._type for col in self._cols]

    @property
    def _names_as_deferred(self) -> list[Deferred | None]:
        return [col._name_as_deferred for col in self._cols]

    def _names_as_deferred_resolved(self, tbl: ir.Table) -> list[ir.Value | None]:
        init = [col._name_as_deferred_resolved(tbl) for col in self._cols]
        return [i for i in init if i is not None]

    def _get_direct_col_sources(self, unique: bool = True) -> list[str]:
        # ibis table, not vinylTable
        col_sources = []
        for col in self._queryable:
            if isinstance(col, Deferred):
                col = col.resolve(self._tbl)
            if callable(col):
                col = col(self._tbl)
            sel_sources = [i.name for i in col.op().find_topmost(ops.Field)]
            if unique:
                col_sources.extend(sel_sources)
            else:
                col_sources.append(sel_sources)

        if unique:
            col_sources = list(set(col_sources))

        return col_sources


class SortColumnListBuilder(ColumnListBuilder):
    _reverse: bool

    def __init__(
        self, tbl, cols, reverse=False, unique=False, passthrough_deferred=False
    ):
        # interpret all sort cols with a - on the outside as descending
        cols = _process_cols_arg(tbl, cols)
        new_cols = []
        for col in cols:
            if col is None:
                new_cols.append(None)
                continue
            col_op = col.op()
            if isinstance(col_op, ops.Negate):
                arg = col_op.args[0]
                new_cols.append(ops.SortKey(arg, ascending=False).to_expr())
            else:
                new_cols.append(col)

        super().__init__(
            tbl, new_cols, unique, passthrough_deferred=passthrough_deferred
        )
        self._reverse = reverse

    @property
    def _sorted(self) -> list[ops.SortKey]:
        final = []
        for so in self._cols:
            # columns can't be None
            if so._col is None:
                continue

            # columns must be resolved for this to work
            if isinstance(so._col, Deferred):
                so._col = so._col.resolve(self._tbl)

            op_it = so._col.op()
            if isinstance(op_it, ops.SortKey):
                # in this case ops_it is a tuple, where the first key is the column and the second is the sort order bool (true is ascending)
                col_expr = op_it.args[0].to_expr()
                out_it = (
                    col_expr.asc() if op_it.args[1] ^ self._reverse else col_expr.desc()
                )
            else:
                out_it = so._col.desc() if self._reverse else so._col.asc()
            final.append(out_it)

        return final

    @property
    def _unsorted(self) -> list[ir.Value]:
        final = []
        for so in self._cols:
            # columns can't be None
            if so._col is None:
                continue

            op_it = so._col.op()
            if isinstance(op_it, ops.SortKey):
                # in this case ops_it is a tuple, where the first key is the column and the second is the sort order bool (true is ascending)
                col_expr = op_it.args[0].to_expr()

            elif isinstance(op_it, ops.Alias) and isinstance(
                op_it.args[0], ops.SortKey
            ):
                col_expr = op_it.args[0].args[0].to_expr().name(op_it.args[1])

            else:
                col_expr = so._col
                # in this case, asc is the default
            final.append(col_expr)

        return final

    @property
    def _names_as_deferred_sorted(self) -> list[Deferred]:
        directions = self._sort_directions
        out = []
        for i, name_def in enumerate(self._names_as_deferred):
            if name_def is None:
                continue
            out.append(name_def.asc() if directions[i] else name_def.desc())
        return out

    def _names_as_deferred_resolved_sorted(self, tbl) -> list[ops.SortKey]:
        directions = self._sort_directions
        out = []
        for i, name_def in enumerate(self._names_as_deferred_resolved(tbl)):
            if name_def is None:
                continue
            out.append(name_def.asc() if directions[i] else name_def.desc())
        return out

    @property
    def _sort_directions(self) -> list[bool]:
        sort_directions = []
        for so in self._cols:
            # columns can't be None
            if so._col is None:
                continue
            op_it = so._col.op()
            if isinstance(op_it, ops.SortKey):
                sort_directions.append(op_it.args[1])
            elif isinstance(op_it, ops.Alias) and isinstance(
                op_it.args[0], ops.SortKey
            ):
                sort_directions.append(op_it.args[0].args[1])
            else:
                sort_directions.append(True)

        return sort_directions


# used for autocomplete for VinylTable and MetricStore classes
class ColumnHelper(
    ir.ArrayColumn,
    ir.BinaryColumn,
    ir.BooleanColumn,
    ir.DecimalColumn,
    ir.FloatingColumn,
    ir.INETColumn,
    ir.IntervalColumn,
    ir.JSONColumn,
    ir.LineStringColumn,
    ir.MACADDRColumn,
    ir.MapColumn,
    ir.MultiLineStringColumn,
    ir.MultiPointColumn,
    ir.MultiPolygonColumn,
    ir.NullColumn,
    ir.PointColumn,
    ir.PolygonColumn,
    ir.StringColumn,
    ir.TimestampColumn,
    ir.IntegerColumn,
    ir.UUIDColumn,
):
    def __init__(self, *args, **kwargs):
        super().__init__(self)



from vinyl.lib.project import Project
from vinyl.lib.query_engine import QueryEngine


def test_query_metric():
    project = Project.bootstrap()
    query_engine = QueryEngine(project)

    df = query_engine._metric(
        store="fare_metrics",
        grain="days=1",
        limit=100,
    ).to_pandas()

    assert len(df.columns) == 3
    assert "total_fare" in df.columns
    assert "avg_fare" in df.columns


def test_query_derive_metric():
    project = Project.bootstrap()
    query_engine = QueryEngine(project)

    df = query_engine._metric(
        store="fare_ratio",
        grain="days=1",
        limit=100,
    ).to_pandas()

    assert len(df.columns) == 4
    assert "total_fare" in df.columns
    assert "avg_fare" in df.columns
    assert "fare_ratio" in df.columns


def test_query_nested_derive_metric():
    project = Project.bootstrap()
    query_engine = QueryEngine(project)

    df = query_engine._metric(
        store="square_fare_ratio",
        grain="days=1",
        limit=100,
    ).to_pandas()

    assert len(df.columns) == 5
    assert "total_fare" in df.columns
    assert "avg_fare" in df.columns
    assert "fare_ratio" in df.columns
    assert "square_fare_ratio" in df.columns


def test_query_metric_dimension():
    project = Project.bootstrap()
    query_engine = QueryEngine(project)

    df = query_engine._metric(
        store="fare_metrics",
        grain="days=1",
        dimensions=["store_and_fwd_flag"],
        limit=100,
    ).to_pandas()

    assert len(df.columns) == 4
    assert "total_fare" in df.columns
    assert "avg_fare" in df.columns


def test_query_metric_from_join():
    project = Project.bootstrap()
    query_engine = QueryEngine(project)

    df = query_engine._metric(
        store="sales_metrics",
        grain="months=1",
        limit=100,
    ).to_pandas()

    assert len(df.columns) == 3
    assert "total_txns" in df.columns
    assert "average_txns" in df.columns



import json
from pathlib import Path

DEFAULT_CREDS_PATH = Path(Path.home(), ".vinyl.json")


class User:
    def __init__(self):
        self.credentials_path = DEFAULT_CREDS_PATH
        if not self.credentials_path.exists():
            raise ValueError("No credentials found. Please login first.")

    @property
    def user_info(self):
        with open(self.credentials_path, "r") as f:
            credentials = json.load(f)
            return {
                "email": credentials["email"],
                "user_id": credentials["user_id"],
            }



import pytest

from vinyl.lib.project import Project


@pytest.mark.skip(
    reason="This test only works when vinyl is hooked up to a dbt project. Now using lineage directly from turntable-cli so not necessary."
)
def test_dbt_lineage():
    project = Project.bootstrap()
    sqlproj = project.get_sql_project(
        ids=["internal_project.models.dbt.base_posthog__person"], parallel=False
    )
    sqlproj.optimize()
    sqlproj.get_lineage()
    assert sqlproj.stitch_lineage() != {}



import threading
from typing import Optional

import typer
from textual.app import ComposeResult
from textual.widgets import DataTable, Footer

from vinyl import Field
from vinyl.cli.events import Event, EventLogger
from vinyl.lib.constants import PreviewHelper
from vinyl.lib.definitions import _load_project_defs
from vinyl.lib.erd import create_erd_app
from vinyl.lib.project import Project
from vinyl.lib.query_engine import QueryEngine
from vinyl.lib.utils.graphics import TurntableTextualApp

preview_cli: typer.Typer = typer.Typer(pretty_exceptions_show_locals=False)


class PreviewTable(TurntableTextualApp):
    def __init__(self, rows, columns, shutdown_seconds: Optional[int] = None):
        self._rows = rows
        self._columns = columns
        self._shutdown_seconds = shutdown_seconds
        super().__init__()

    def compose(self) -> ComposeResult:
        yield DataTable()
        yield Footer()

    def on_mount(self) -> None:
        table = self.query_one(DataTable)
        table.add_columns(*self._columns)
        table.add_rows(self._rows)
        if self._shutdown_seconds:
            threading.Timer(self._shutdown_seconds, self.quit).start()

    def quit(self):
        self.exit()


@preview_cli.command("model")
def preview_model(
    name: str = typer.Option(
        False, "--name", "-m", help="exported name of the model or source"
    ),
    twin: bool = typer.Option(False, "--twin", "-t", help="use twin data"),
    limit: int = typer.Option(1000, "--limit", "-l", help="limit the number of rows"),
    shutdown_seconds: Optional[int] = typer.Option(
        None, "--shutdown", "-s", help="shutdown after n seconds. Default is None."
    ),
):
    """Preview a model"""
    track = EventLogger()
    track.log_event(Event.PREVIEW_MODEL)
    if twin:
        PreviewHelper._preview = "twin"

    project = Project.bootstrap()
    query_engine = QueryEngine(project)
    if isinstance(limit, typer.models.OptionInfo):
        # handle typing of default value for limit
        if limit.default is None:
            raise ValueError("limit must have a default that is an integer")
        limit = limit.default
    result = query_engine._model(name, limit=limit)
    app = PreviewTable(
        columns=tuple(result.columns()),
        rows=result.numpy_rows(),
        shutdown_seconds=shutdown_seconds,
    )
    app.run()


@preview_cli.command("metric")
def preview_metric(
    name: str = typer.Option(..., help="exported name of the metric column"),
    grain: str = typer.Option(
        ..., help="grain to bucket the metric by (ex: months=3, year=1)"
    ),
    dims: list[str] = typer.Option(
        None, "--dims", help="comma separated list of dimensions to include"
    ),
    cols: list[str] = typer.Option(
        None, "--cols", help="comma separated list of columns to include"
    ),
    shutdown_seconds: Optional[int] = typer.Option(
        None, "--shutdown", "-s", help="shutdown after n seconds. Default is None."
    ),
):
    """Preview a model"""
    track = EventLogger()
    track.log_event(Event.PREVIEW_METRIC)
    project = Project.bootstrap()
    query_engine = QueryEngine(project)
    result = query_engine._metric(store=name, grain=grain, limit=1000)

    app = PreviewTable(
        columns=result.columns(),
        rows=result.numpy_rows(),
        shutdown_seconds=shutdown_seconds,
    )
    app.run()


@preview_cli.command("erd")
def erd(
    names: list[str] = typer.Option(
        [], "--name", "-m", help="exported name(s) of the model or source"
    ),
):
    """Generate an ERD"""
    _load_project_defs()
    G = Field._export_relations_to_networkx(
        shorten_name=True, filter=None if len(names) == 0 else names
    )
    create_erd_app(G).run()



import orjson

from vinyl.lib.errors import VinylError


class ValidationOutput:
    errors: list[VinylError] | None
    bytes_processed: float | None
    cost: float | None

    def __init__(
        self,
        errors: list[VinylError] | None,
        bytes_processed: float | None,
        cost: float | None,
    ):
        self.errors = errors
        self.bytes_processed = bytes_processed
        self.cost = cost

    def to_dict(self):
        return {
            "errors": [error.to_dict() for error in self.errors]
            if self.errors
            else None,
            "bytes_processed": self.bytes_processed,
            "cost": self.cost,
        }

    def to_json(self):
        return orjson.dumps(self.to_dict())



# type: ignore
from internal_project.resources import local_filesystem  # noqa F401
from internal_project.sources.local_filesystem.stores import Stores
from vinyl import Field, source  # noqa F401
from vinyl import types as t  # noqa F401


@source(resource=local_filesystem)
class Test:
    _table = "test"
    _unique_name = "local_filesystem.Test"
    _path = "data/csvs/test.csv"
    _row_count = 28512
    _col_replace = {}

    id: t.Int64(nullable=True) = Field(primary_key=True)
    date: t.Date(nullable=True)
    store_nbr: t.Int64(nullable=True) = Field(foreign_key=Stores.store_nbr)
    family: t.String(nullable=True)
    onpromotion: t.Int64(nullable=True)



from vinyl.lib.utils.files import _get_changed_files


def test_generate_sources():
    @_get_changed_files("internal_project/sources")
    def test():
        import vinyl.cli.sources as sources_cli

        sources_cli.generate_sources(twin=False, resources=["local_filesystem"])

    assert test() == {}


def test_preview_model():
    import vinyl.cli.preview as preview_cli

    preview_cli.preview_model(name="amount_base", twin=False, shutdown_seconds=1)

    assert True


def test_preview_metric():
    import vinyl.cli.preview as preview_cli

    preview_cli.preview_metric(name="fare_metrics", grain="days=1", shutdown_seconds=1)

    assert True



import json
import os

from vinyl import load_dotenv
from vinyl.lib.asset import resource
from vinyl.lib.connect import (
    BigQueryConnector,
    DatabaseFileConnector,
    FileConnector,
    PostgresConnector,
    SnowflakeConnector,
)

load_dotenv()


@resource
def local_filesystem():
    return FileConnector(path="data/csvs")


@resource
def local_parquet():
    return FileConnector(path="data/parquets")


@resource
def local_json():
    return FileConnector(path="data/jsons")


@resource
def local_duckdb():
    return DatabaseFileConnector(path="../tests/vinyl/fixtures/test.duckdb")


@resource
def supabase_copilot_postgres():
    return PostgresConnector(
        host="aws-0-us-east-1.pooler.supabase.com",
        port="5432",
        user=os.getenv("SUPABASE_COPILOT_USER"),
        password=os.getenv("SUPABASE_COPILOT_PASSWORD"),
        tables=["postgres.public.*"],
    )


@resource
def dbt_bigquery():
    return BigQueryConnector(
        service_account_info=json.loads(os.getenv("GOOGLE_APPLICATION_CREDENTIALS")),
        tables=["analytics-dev-372514.dbt_itracey.companies"],
    )


@resource
def dbt_snowflake():
    return SnowflakeConnector(
        account=os.getenv("SNOWFLAKE_ACCOUNT"),
        user=os.getenv("SNOWFLAKE_USER"),
        password=os.getenv("SNOWFLAKE_PASSWORD"),
        tables=["ANALYTICS.MIXPANEL.EVENT"],
    )


# @resource
# def dbt_turntable_dbt():
#     return DBTConnectorFull(
#         dbt_project_dir="../../turntable_dbt/",
#         dialect=DBTDialect.BIGQUERY,
#         version=DBTVersion.V1_5,
#         tables=[
#             "analytics-dev-372514.plausible.stats_timeseries_results",
#             "analytics-dev-372514.linkedin_company_pages.comment_history",
#             "analytics-dev-372514.linkedin_company_pages.comment_on_ugc_post",
#             "analytics-dev-372514.linkedin_company_pages.comment_social_metadata_summary",
#             "analytics-dev-372514.linkedin_company_pages.country",
#             "analytics-dev-372514.linkedin_company_pages.followers_by_association_type",
#             "analytics-dev-372514.linkedin_company_pages.followers_by_function",
#             "analytics-dev-372514.linkedin_company_pages.followers_by_industry",
#             "analytics-dev-372514.linkedin_company_pages.followers_by_seniority",
#             "analytics-dev-372514.linkedin_company_pages.followers_by_staff_count_range",
#             "analytics-dev-372514.linkedin_company_pages.function",
#             "analytics-dev-372514.linkedin_company_pages.industry",
#             "analytics-dev-372514.linkedin_company_pages.organization",
#             "analytics-dev-372514.linkedin_company_pages.organization_followers_by_association_type",
#             "analytics-dev-372514.linkedin_company_pages.organization_followers_by_function",
#             "analytics-dev-372514.linkedin_company_pages.organization_followers_by_industry",
#             "analytics-dev-372514.linkedin_company_pages.organization_followers_by_seniority",
#             "analytics-dev-372514.linkedin_company_pages.organization_followers_by_staff_count_range",
#             "analytics-dev-372514.linkedin_company_pages.organization_industries",
#             "analytics-dev-372514.linkedin_company_pages.organization_page_statistic_by_country",
#             "analytics-dev-372514.linkedin_company_pages.organization_page_statistic_by_function",
#             "analytics-dev-372514.linkedin_company_pages.organization_page_statistic_by_industry",
#             "analytics-dev-372514.linkedin_company_pages.organization_page_statistic_by_region",
#             "analytics-dev-372514.linkedin_company_pages.organization_page_statistic_by_seniority",
#             "analytics-dev-372514.linkedin_company_pages.organization_page_statistic_by_staff_count_range",
#             "analytics-dev-372514.linkedin_company_pages.organization_share",
#             "analytics-dev-372514.linkedin_company_pages.organization_time_bound_follower_statistic",
#             "analytics-dev-372514.linkedin_company_pages.organization_time_bound_page_statistic",
#             "analytics-dev-372514.linkedin_company_pages.organization_time_bound_share_statistic",
#             "analytics-dev-372514.linkedin_company_pages.organization_total_follower_statistic",
#             "analytics-dev-372514.linkedin_company_pages.organization_total_page_statistic",
#             "analytics-dev-372514.linkedin_company_pages.organization_total_share_statistic",
#             "analytics-dev-372514.linkedin_company_pages.page_statistic_by_country",
#             "analytics-dev-372514.linkedin_company_pages.page_statistic_by_function",
#             "analytics-dev-372514.linkedin_company_pages.page_statistic_by_industry",
#             "analytics-dev-372514.linkedin_company_pages.page_statistic_by_region",
#             "analytics-dev-372514.linkedin_company_pages.page_statistic_by_seniority",
#             "analytics-dev-372514.linkedin_company_pages.page_statistic_by_staff_count_range",
#             "analytics-dev-372514.linkedin_company_pages.region",
#             "analytics-dev-372514.linkedin_company_pages.seniority",
#             "analytics-dev-372514.linkedin_company_pages.share_content",
#             "analytics-dev-372514.linkedin_company_pages.share_history",
#             "analytics-dev-372514.linkedin_company_pages.share_share_statistic",
#             "analytics-dev-372514.linkedin_company_pages.share_statistic",
#             "analytics-dev-372514.linkedin_company_pages.social_metadata_summary",
#             "analytics-dev-372514.linkedin_company_pages.time_bound_follower_statistic",
#             "analytics-dev-372514.linkedin_company_pages.time_bound_page_statistic",
#             "analytics-dev-372514.linkedin_company_pages.time_bound_share_statistic",
#             "analytics-dev-372514.linkedin_company_pages.total_follower_statistic",
#             "analytics-dev-372514.linkedin_company_pages.total_page_statistic",
#             "analytics-dev-372514.linkedin_company_pages.total_share_statistic",
#             "analytics-dev-372514.linkedin_company_pages.ugc_post_social_metadata_summary",
#             "analytics-dev-372514.pipedrive.activity",
#             "analytics-dev-372514.pipedrive.activity_participant",
#             "analytics-dev-372514.pipedrive.activity_type",
#             "analytics-dev-372514.pipedrive.currency",
#             "analytics-dev-372514.pipedrive.deal_history",
#             "analytics-dev-372514.pipedrive.email_bcc",
#             "analytics-dev-372514.pipedrive.email_cc",
#             "analytics-dev-372514.pipedrive.email_to",
#             "analytics-dev-372514.pipedrive.file",
#             "analytics-dev-372514.pipedrive.filter",
#             "analytics-dev-372514.pipedrive.filter_helper",
#             "analytics-dev-372514.pipedrive.fivetran_audit",
#             "analytics-dev-372514.pipedrive.lead_history",
#             "analytics-dev-372514.pipedrive.lead_label_history",
#             "analytics-dev-372514.pipedrive.lead_lead_label",
#             "analytics-dev-372514.pipedrive.lead_source",
#             "analytics-dev-372514.pipedrive.mail_message",
#             "analytics-dev-372514.pipedrive.mail_thread",
#             "analytics-dev-372514.pipedrive.note",
#             "analytics-dev-372514.pipedrive.organization",
#             "analytics-dev-372514.pipedrive.person",
#             "analytics-dev-372514.pipedrive.person_email",
#             "analytics-dev-372514.pipedrive.person_phone",
#             "analytics-dev-372514.pipedrive.pipeline",
#             "analytics-dev-372514.pipedrive.product",
#             "analytics-dev-372514.pipedrive.product_deal",
#             "analytics-dev-372514.pipedrive.role",
#             "analytics-dev-372514.pipedrive.stage",
#             "analytics-dev-372514.pipedrive.user",
#             "analytics-dev-372514.linear.issues",
#             "analytics-dev-372514.quickbooks_quickbooks.quickbooks__balance_sheet",
#             "analytics-dev-372514.quickbooks_quickbooks.quickbooks__expenses_sales_enhanced",
#             "analytics-dev-372514.quickbooks_quickbooks.quickbooks__general_ledger",
#             "analytics-dev-372514.quickbooks_quickbooks.quickbooks__general_ledger_by_period",
#             "analytics-dev-372514.quickbooks_quickbooks.quickbooks__profit_and_loss",
#             "analytics-dev-372514.mixpanel.event",
#             "analytics-dev-372514.mixpanel.fivetran_audit",
#             "analytics-dev-372514.mixpanel.people",
#             "analytics-dev-372514.supabase_public.inferences",
#             "analytics-dev-372514.supabase_public.responses",
#             "analytics-dev-372514.supabase_public.waitlist",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__companies",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__company_history",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__contact_history",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__contact_lists",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__contacts",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__deal_history",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__deal_stages",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__deals",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__engagement_calls",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__engagement_emails",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__engagement_meetings",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__engagement_notes",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__engagement_tasks",
#             "analytics-dev-372514.hubspot_hubspot.hubspot__engagements",
#             "analytics-dev-372514.hubspot_hubspot.int_hubspot__contact_merge_adjust",
#             "analytics-dev-372514.github.asset",
#             "analytics-dev-372514.github.branch_commit_relation",
#             "analytics-dev-372514.github.commit",
#             "analytics-dev-372514.github.commit_file",
#             "analytics-dev-372514.github.commit_parent",
#             "analytics-dev-372514.github.commit_pull_request",
#             "analytics-dev-372514.github.deployment",
#             "analytics-dev-372514.github.deployment_status",
#             "analytics-dev-372514.github.github__daily_metrics",
#             "analytics-dev-372514.github.github__issues",
#             "analytics-dev-372514.github.github__monthly_metrics",
#             "analytics-dev-372514.github.github__pull_requests",
#             "analytics-dev-372514.github.github__quarterly_metrics",
#             "analytics-dev-372514.github.github__weekly_metrics",
#             "analytics-dev-372514.github.issue",
#             "analytics-dev-372514.github.issue_assignee",
#             "analytics-dev-372514.github.issue_assignee_history",
#             "analytics-dev-372514.github.issue_closed_history",
#             "analytics-dev-372514.github.issue_comment",
#             "analytics-dev-372514.github.issue_label",
#             "analytics-dev-372514.github.issue_label_history",
#             "analytics-dev-372514.github.issue_mention",
#             "analytics-dev-372514.github.issue_merged",
#             "analytics-dev-372514.github.issue_referenced",
#             "analytics-dev-372514.github.issue_renamed",
#             "analytics-dev-372514.github.label",
#             "analytics-dev-372514.github.pull_request",
#             "analytics-dev-372514.github.pull_request_ready_for_review_history",
#             "analytics-dev-372514.github.pull_request_review",
#             "analytics-dev-372514.github.pull_request_review_comments",
#             "analytics-dev-372514.github.release",
#             "analytics-dev-372514.github.repository",
#             "analytics-dev-372514.github.requested_reviewer_history",
#             "analytics-dev-372514.github.security_advisory",
#             "analytics-dev-372514.github.security_advisory_cwe",
#             "analytics-dev-372514.github.security_advisory_vulnerability",
#             "analytics-dev-372514.github.security_alert",
#             "analytics-dev-372514.github.security_cwe",
#             "analytics-dev-372514.github.security_reference",
#             "analytics-dev-372514.github.security_vulnerability",
#             "analytics-dev-372514.github.stg_github__issue",
#             "analytics-dev-372514.github.stg_github__issue_assignee",
#             "analytics-dev-372514.github.stg_github__issue_assignee_tmp",
#             "analytics-dev-372514.github.stg_github__issue_closed_history",
#             "analytics-dev-372514.github.stg_github__issue_closed_history_tmp",
#             "analytics-dev-372514.github.stg_github__issue_comment",
#             "analytics-dev-372514.github.stg_github__issue_comment_tmp",
#             "analytics-dev-372514.github.stg_github__issue_label",
#             "analytics-dev-372514.github.stg_github__issue_label_tmp",
#             "analytics-dev-372514.github.stg_github__issue_merged",
#             "analytics-dev-372514.github.stg_github__issue_merged_tmp",
#             "analytics-dev-372514.github.stg_github__issue_tmp",
#             "analytics-dev-372514.github.stg_github__label",
#             "analytics-dev-372514.github.stg_github__label_tmp",
#             "analytics-dev-372514.github.stg_github__pull_request",
#             "analytics-dev-372514.github.stg_github__pull_request_review",
#             "analytics-dev-372514.github.stg_github__pull_request_review_tmp",
#             "analytics-dev-372514.github.stg_github__pull_request_tmp",
#             "analytics-dev-372514.github.stg_github__repository",
#             "analytics-dev-372514.github.stg_github__repository_tmp",
#             "analytics-dev-372514.github.stg_github__requested_reviewer_history",
#             "analytics-dev-372514.github.stg_github__requested_reviewer_history_tmp",
#             "analytics-dev-372514.github.stg_github__user",
#             "analytics-dev-372514.github.stg_github__user_tmp",
#             "analytics-dev-372514.github.user",
#             "analytics-dev-372514.github.user_email",
#             "analytics-dev-372514.github.workflow",
#             "analytics-dev-372514.github.workflow_run",
#             "analytics-dev-372514.github.workflow_run_pull_request",
#             "analytics-dev-372514.posthog.data_attribute",
#             "analytics-dev-372514.posthog.event",
#             "analytics-dev-372514.posthog.insight",
#             "analytics-dev-372514.posthog.person",
#             "analytics-dev-372514.posthog.person_name",
#             "analytics-dev-372514.posthog.project",
#             "analytics-dev-372514.posthog.test_account_filter",
#             "analytics-dev-372514.posthog.users",
#         ],
#     )



from vinyl.lib.constructors import (  # noqa: F401
    date,
    interval,
    map,
    now,
    random,
    struct,
    time,
    timestamp,
)
from vinyl.lib.constructors import literal as lit  # noqa: F401



from __future__ import annotations

import ast
import os
from typing import Any

from core.db_methods import (
    dbt_repo_context,
)
from vinyl import load_dotenv
from supabase import Client

from backend.core.serialization import (  # noqa
    ModelObject,
    ResourceObject,
    SourceObject,
)
from vinyl.cli.dbt import (
    DBTProjectStructure,
    generate_models_helper,
    generate_resources_helper,
)
from vinyl.cli.sources import generate_sources_helper
from vinyl.lib.dbt import DBTDialect
from vinyl.lib.dbt_methods import DBTVersion
from vinyl.lib.definitions import Defs, _create_dag_from_model_dict_list  # noqa
from vinyl.lib.project import Project

load_dotenv()

ENCRYPTION_KEY = os.getenv("ENCRYPTION_KEY")


def undo_insertion_on_error(func):
    def wrapper(self: VinylDB, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except Exception as e:
            self.undo_insertion_on_error_helper()
            self.errors_cleaned = True
            raise e

    return wrapper


class VinylDB:
    def __init__(self, supabase_client: Client, tenant_id: str | None = None):
        self.supabase_client = supabase_client
        self.tenant_id = tenant_id
        self.resource_id_dict = {}
        self.asset_id_dict = {}
        self.asset_link_id_dict = {}
        self.column_id_dict = {}
        self.column_link_id_dict = {}
        self.resource_id_dict = {}
        self.errors_cleaned = (
            False  # ensures that errors are only cleaned once even for nested calls
        )

    @undo_insertion_on_error
    def setup_dbt_project(
        self,
        tenant_id: str,
        repo_id: str,
        dialect: DBTDialect,
        version: DBTVersion,
        gen_models: bool = False,
    ):
        with dbt_repo_context(
            self.supabase_client, repo_id, tenant_id, dialect=dialect, version=version
        ) as dbtproj:
            res = generate_resources_helper(dbtproj, json_=True)
            res["repository_id"] = repo_id
            out = {}
            data, count = self.supabase_client.table("resources").insert(res).execute()
            out["resources"] = self.update_resources_data(data[-1])
            self.resource_id_dict = {v["name"]: v["id"] for v in out["resources"]}
            if gen_models:
                models, asset_links = generate_models_helper(
                    dbtproj,
                    structure=DBTProjectStructure.JSON,
                )
                out["models"] = models
                out["asset_links"] = asset_links

            return out

    def undo_insertion_on_error_helper(self):
        dict_map = {
            "column_links": self.column_link_id_dict,
            "columns": self.column_id_dict,
            "asset_links": self.asset_link_id_dict,
            "assets": self.asset_id_dict,
            "resources": self.resource_id_dict,
        }  # order of dict map is made to minimize fk constraint error risk
        for k, v in dict_map.items():
            if len(v) > 0 and not self.errors_cleaned:
                data, count = (
                    self.supabase_client.table(k)
                    .delete()
                    .in_("id", [v for v in v.values()])
                    .execute()
                )

    def update_resources_data(self, resources_data):
        for v in resources_data:
            # make sure input is a list not a jsonified string
            if isinstance(v["tables"], str):
                v["tables"] = ast.literal_eval(v["tables"])
            v["supabase_url"] = self.url
            v["supabase_key"] = self.key
        return resources_data

    @undo_insertion_on_error
    def generate_sources(self, resource_ids: list[str]):
        defs_dict = {}
        data, count = (
            self.supabase_client.table("resources")
            .select("*")
            .in_("id", resource_ids)
            .execute()
        )
        resources_data = self.update_resources_data(data[-1])
        self.resource_id_dict = {v["name"]: v["id"] for v in resources_data}

        defs_dict["resources"] = resources_data

        defs = Defs.load_from_dict(defs_dict)

        sources, source_json = generate_sources_helper(json=True, defs=defs, twin=False)
        for v in source_json:
            v["resource_ids"] = [self.resource_id_dict[v["resource_name"]]]
        return source_json

    @undo_insertion_on_error
    def ingest_dbt_project(
        self, repo_id: str, tenant_id: str, dialect: DBTDialect, version: DBTVersion
    ):
        dbt_proj_info = self.setup_dbt_project(
            repo_id, dialect, tenant_id, version, gen_models=True
        )
        sources = self.generate_sources([v["id"] for v in dbt_proj_info["resources"]])
        model_dag = _create_dag_from_model_dict_list(dbt_proj_info["models"])
        for model in dbt_proj_info["models"]:
            ancestor_roots = model_dag.get_ancestor_roots([model["unique_name"]])
            source_ancestor_roots = [n for n in ancestor_roots if ".sources." in n]
            resource_ids = []
            for v in sources:
                if v["unique_name"] in source_ancestor_roots:
                    resource_ids.append(self.resource_id_dict[v["resource_name"]])

            # handle sql models where model may not be connected to a source
            if resource_ids == []:
                res_name = model["config"]["steps"][0].get("resource_name", None)
                if res_name is not None:
                    resource_ids.append(self.resource_id_dict[res_name])
            model["resource_ids"] = list(set(resource_ids))
        for source in sources:
            del source["resource_name"]  # no longer needed

        defs_dict = {
            "resources": dbt_proj_info["resources"],
            "sources": sources,
            "models": dbt_proj_info["models"],
        }
        self.upload_assets(defs_dict, dbt_proj_info["asset_links"])
        return Project.bootstrap_from_dict(defs_dict)

    @undo_insertion_on_error
    def upload_assets(
        self, defs_dict: dict[str, Any], asset_links: list[dict[str, str]]
    ):
        assets_to_upload = [
            SourceObject(**i).model_dump()  # type: ignore[attr-defined]
            for i in defs_dict["sources"]
        ]
        assets_to_upload += [ModelObject(**i).model_dump() for i in defs_dict["models"]]  # type: ignore[attr-defined]
        data, count = (
            self.supabase_client.table("assets").insert(assets_to_upload).execute()
        )
        self.asset_id_dict = {v["unique_name"]: v["id"] for v in data[1]}
        adj_asset_links = []
        for link in asset_links:
            adj_asset_links.append(
                {
                    "type": link["type"],
                    "source_id": self.asset_id_dict[link["source_unique_name"]],
                    "target_id": self.asset_id_dict[link["target_unique_name"]],
                }
            )
        data, count = (
            self.supabase_client.table("asset_links").insert(adj_asset_links).execute()
        )
        self.asset_link_id_dict = {v["source_id"]: v["id"] for v in data[1]}
        return data[-1]

    @undo_insertion_on_error
    def upload_columns(
        self,
        resource_ids: list[str] | None = None,
        ids: list[str] | None = None,
        predecessor_depth: int | None = None,
        successor_depth: int | None = None,
    ):
        project = self.get_project(resource_ids)
        sqlproject = project.get_sql_project(
            parallel=False,
            ids=ids,
            predecessor_depth=predecessor_depth,
            successor_depth=successor_depth,
        )
        sqlproject.optimize()
        columns, column_links = sqlproject.stitch_lineage_postgres()
        assets_data = self.get_assets_data(resource_ids)
        if self.asset_id_dict == {}:
            if resource_ids is None:
                raise ValueError(
                    "resource_ids must be provided if asset_id_dict is empty"
                )
            # in case running as a standalone function, get correct assets but dont update asset_id_dict
            asset_id_dict = {v["unique_name"]: v["id"] for v in assets_data}
        else:
            asset_id_dict = self.asset_id_dict
        for col in columns:
            col["asset_id"] = asset_id_dict.get(col["table_unique_name"], None)
            if "errors" in col and col["errors"] != []:
                col["errors"] = [err.to_dict() for err in col["errors"]]
            if col["asset_id"] is not None:
                asset_it = [v for v in assets_data if v["id"] == col["asset_id"]][0]
                col["description"] = (
                    asset_it["config"]
                    .get("column_descriptions", {})
                    .get(col["name"], None)
                )
        data, count = self.supabase_client.table("columns").insert(columns).execute()
        self.column_id_dict = {
            f"{v['table_unique_name']}.{v['name']}": v["id"] for v in data[1]
        }
        adj_column_links = []
        for link in column_links:
            adj_column_links.append(
                {
                    "lineage_type": link["lineage_type"],
                    "connection_types": link["connection_types"],
                    "source_id": self.column_id_dict[link["source_name"]],
                    "target_id": self.column_id_dict[link["target_name"]],
                }
            )
        data, count = (
            self.supabase_client.table("column_links")
            .insert(adj_column_links)
            .execute()
        )
        self.column_link_id_dict = {v["source_id"]: v["id"] for v in data[1]}

    def get_resources_data(
        self, repo_ids: list[str] = [], additional_resources: list[str] = []
    ):
        builder = self.supabase_client.table("resources")
        query = "id, name, type, tables, config, profiles(encrypted_secret)"
        if repo_ids:
            query += ", repositories(id, tables, repo_path_zip)"
        builder = builder.select(query)
        if repo_ids and additional_resources:
            repo_id_string = ", ".join(repo_ids)
            resource_id_string = ", ".join(additional_resources)
            builder = builder.or_(
                f"repositories.id.in.{repo_id_string}",
                f"id.in.{resource_id_string}",
            )
        elif repo_ids:
            builder = builder.in_("repositories.id", repo_ids)
        elif additional_resources:
            builder = builder.in_("id", additional_resources)
        else:
            raise ValueError("Must provide either repo_ids or additional_resources")

        if self.tenant_id is not None:
            builder.eq("tenant_id", self.tenant_id)

        data, count = builder.execute()
        resources = data[-1]
        for res in resources:
            repositories = res.get("repositories", None)
            if repositories is not None:
                if len(repositories) > 1:
                    raise ValueError(
                        "Multiple repositories found for a resource after filtering"
                    )
                if len(repositories) == 1:
                    res["repositories"] = repositories[-1]

        return resources

        # rename resource fields to be compatible with ResourceObject
        # for res in resources:
        #     if "repositories" in res:
        #         res["repository_config"] = res.pop("repositories")
        #     if "profiles" in res:
        #         res["profiles_config"] = res.pop("profiles")

        return resources

    def get_assets_data(self, resource_ids: list[str] | None = None):
        # get sources
        builder = self.supabase_client.table("assets").select("*")
        if resource_ids is None:
            builder = builder.is_("resource_ids", "null")
        else:
            stringified = "{"
            for i, v in enumerate(resource_ids):
                stringified += f'"{v}"'
                if i != len(resource_ids) - 1:
                    stringified += ","
            stringified += "}"
            builder = builder.filter("resource_ids", "cs", stringified)
        data, count = builder.execute()

        return data[-1]

    def get_defs_dict(self, resource_ids: list[str] | None = None):
        resources_data = self.get_resources_data(resource_ids)
        for v in resources_data:
            v["supabase_url"] = self.url
            v["supabase_key"] = self.key
        assets_data = self.get_assets_data(resource_ids)
        models_data = [v for v in assets_data if v["type"] == "model"]
        sources_data = [v for v in assets_data if v["type"] == "source"]
        return {
            "resources": resources_data,
            "sources": sources_data,
            "models": models_data,
        }

    def get_project(self, resource_ids: list[str] | None = None):
        defs_dict = self.get_defs_dict(resource_ids)
        return Project.bootstrap_from_dict(defs_dict)

    def get_related_assets(
        self,
        asset_ids: list[str] | None = None,
        predecessor_depth: int | None = None,
        successor_depth: int | None = None,
    ):
        successor_data, count = self.supabase_client.rpc(
            "assets_bfs",
            {
                "nodes": asset_ids,
                "max_depth": successor_depth,
            },
        ).execute()
        predecessor_data, count = self.supabase_client.rpc(
            "assets_bfs_inverse",
            {
                "nodes": asset_ids,
                "max_depth": predecessor_depth,
            },
        ).execute()

        asset_ids.extend([v["node_id"] for v in predecessor_data[1]])
        asset_ids.extend([v["node_id"] for v in successor_data[1]])
        return asset_ids

    async def get_lineage(
        self,
        asset_id: str,
    ):
        column_data, count = (
            self.supabase_client.table("columns")
            .select("*")
            .in_("asset_id", asset_id)
            .execute()
        )
        column_ids = [v["id"] for v in column_data[-1]]
        column_links_data, count = (
            self.supabase_cliendt.table("column_links")
            .select("*")
            .in_("target_id", column_ids)
            .execute()
        )
        return column_data[-1], column_links_data[-1]

    def clean_db(self):
        for tbl in ["resources", "assets", "columns", "column_links", "asset_links"]:
            self.supabase_client.table(tbl).delete().filter(
                "id", "not.is", "null"
            ).execute()


if __name__ == "__main__":
    from vinyl import load_dotenv

    load_dotenv()
    from datastore import supabase

    db = VinylDB(supabase_client=supabase)
    # x = db.upload_columns(
    #     resource_ids=["46045b81-829d-4698-a280-b87a6e708814"],
    #     ids=["stg_posthog__events"],
    # )
    y = db.get_resources_data(repo_ids=["b7b5d42e-6558-410c-ac7b-c8e116a6c802"])[0]
    from backend.core.serialization import ResourceObject

    yy = ResourceObject(**y)
    yy._get_connector()._connect()
    # breakpoint()
    # x = get_dbt_profile_contents_from_resource(data[1])
    # import os
    # from base64 import urlsafe_b64decode

    # from cryptography.fernet import Fernet
    # from vinyl import load_dotenv

    # load_dotenv()

    # ENCRYPTION_KEY = os.getenv("ENCRYPTION_KEY")
    # cipher = Fernet(ENCRYPTION_KEY)

    # decrypted_secret = cipher.decrypt(urlsafe_b64decode(secret)).decode("utf-8")
    breakpoint()
    # db.clean_db()
    # proj = db.ingest_dbt_project(
    #     "8200dcbe-0641-4a66-9091-ce7482319603", DBTDialect.BIGQUERY, DBTVersion.V1_5
    # )
    # print(db.get_assets_data(resource_ids=["0e30d4e2-76c4-4538-967c-0491551e1452"]))
    # db.upload_columns(
    #     resource_ids=["46045b81-829d-4698-a280-b87a6e708814"],
    #     # ids=["internal_project.models.dbt.base_posthog__users"],
    # )
